(******************************************************************************)
(*  ######  #####  #####  ###   ###    A Formal C Concurrency Model           *)
(*  ##     ##     ##      #### ####                                           *)
(*  #####  ##     ##      ## ### ##    Copyright (c) Viktor Vafeiadis         *)
(*  ##     ##     ##      ##     ##                                           *)
(*  ##      #####  #####  ##     ##    See LICENSE.txt for license.           *)
(******************************************************************************)

Require Import Vbase List Relations. 
Set Implicit Arguments.

Lemma clos_refl_transE A r a b :
  clos_refl_trans A r a b <-> a = b \/ clos_trans A r a b.
Proof.
  split; ins; desf; vauto; induction H; desf; vauto. 
Qed.

Lemma clos_trans_in_rt A r a b :
  clos_trans A r a b -> clos_refl_trans A r a b.
Proof.
  induction 1; vauto. 
Qed.

Lemma rt_t_trans A r a b c :
  clos_refl_trans A r a b -> clos_trans A r b c -> clos_trans A r a c.
Proof.
  ins; induction H; eauto using clos_trans. 
Qed.

Lemma t_rt_trans A r a b c :
  clos_trans A r a b -> clos_refl_trans A r b c -> clos_trans A r a c.
Proof.
  ins; induction H0; eauto using clos_trans. 
Qed.

Lemma clos_trans_mon A (r r' : relation A) a b :
  clos_trans A r a b ->
  (forall a b, r a b -> r' a b) ->
  clos_trans A r' a b.
Proof.
  induction 1; ins; eauto using clos_trans.
Qed.

Lemma clos_refl_trans_mon A (r r' : relation A) a b :
  clos_refl_trans A r a b ->
  (forall a b, r a b -> r' a b) ->
  clos_refl_trans A r' a b.
Proof.
  induction 1; ins; eauto using clos_refl_trans.
Qed.

Lemma t_step_rt A R x y : 
  clos_trans A R x y <-> exists z, R x z /\ clos_refl_trans A R z y.
Proof.
  split; ins; desf.
    by apply clos_trans_tn1 in H; induction H; desf; eauto using clos_refl_trans.
  by rewrite clos_refl_transE in *; desf; eauto using clos_trans.
Qed.


(******************************************************************************)
(** * C11 concurrency model definition *)
(******************************************************************************)

Definition immediate X (rel : relation X) (a b: X) :=
  rel a b /\ (forall c (R1: rel a c) (R2: rel c b), False).

Definition irreflexive X (rel : relation X) := forall x, rel x x -> False.

Definition acyclic X (rel : relation X) := irreflexive (clos_trans X rel).

Definition is_total X (cond: X -> Prop) (rel: relation X) :=
  forall a (IWa: cond a)
         b (IWb: cond b) (NEQ: a <> b),
    rel a b \/ rel b a.

Definition restr_subset X (cond: X -> Prop) (rel rel': relation X) :=
  forall a (IWa: cond a)
         b (IWb: cond b) (REL: rel a b),
    rel' a b.

Definition actid := nat.

Inductive read_access_type := 
  | RATsc
  | RATacq 
  | RATcons
  | RATrlx 
  | RATna.

Inductive write_access_type := 
  | WATsc
  | WATrel
  | WATrlx 
  | WATna.

Inductive rmw_access_type := 
  | UATsc
  | UATacq_rel
  | UATcons_rel (* not in C11 *)
  | UATacq 
  | UATrel
  | UATcons
  | UATrlx.

Inductive action := 
  | Askip 
  | Aalloc (l n : nat)
  | Aload (typ : read_access_type) (l : nat) (v : nat) 
  | Astore (typ : write_access_type) (l : nat) (v : nat) 
  | Armw (typ : rmw_access_type) (l : nat) (f : nat -> nat) (v : nat).

Definition loc a :=
  match a with
    | Askip => 0
    | Aalloc l _
    | Aload _ l _ 
    | Astore _ l _
    | Armw _ l _ _ => l
  end. 

Definition is_rmw a :=
  match a with
    | Askip | Aalloc _ _ | Aload _ _ _ | Astore _ _ _ => false 
    | Armw _ _ _ _ => true
  end.

Definition is_write a :=
  match a with
    | Askip | Aalloc _ _ | Aload _ _ _ => false 
    | Astore _ _ _ | Armw _ _ _ _ => true
  end.
 
Definition is_access a :=
  match a with
    | Askip | Aalloc _ _ => false 
    | Aload _ _ _  | Astore _ _ _ | Armw _ _ _ _ => true
  end. 

Definition is_accessL a l :=
  match a with
    | Aload _ l' _ 
    | Astore _ l' _ 
    | Armw _ l' _ _ => l' = l
    | _ => False
  end.

Definition is_writeL a l :=
  match a with
    | Astore _ l' _ 
    | Armw _ l' _ _ => l' = l
    | _ => False
  end.

Definition is_writeLV a l v :=
  match a with
    | Astore _ l' v' => l' = l /\ v' = v 
    | Armw _ l' f v' => l' = l /\ f v' = v
    | _ => False
  end.

Definition is_readL a l :=
  match a with
    | Aload _ l' _ 
    | Armw _ l' _ _ => l' = l
    | _ => False
  end.

Definition is_readLV a l v :=
  match a with
    | Aload _ l' v' 
    | Armw _ l' _ v' => l' = l /\ v' = v
    | _ => False
  end.

Definition is_only_read a :=
  match a with
    | Aload _ _ _ => True
    | _ => False
  end.

Definition is_only_readL a l :=
  match a with
    | Aload _ l' _ => l' = l
    | _ => False
  end.

Definition is_na a :=
  match a with
    | Aload RATna _ _  
    | Astore WATna _ _ => true
    | _ => false
  end.

(** Is the action sequentially consistent? *)

Definition is_sc a :=
  match a with
    | Aload RATsc _ _  
    | Astore WATsc _ _  
    | Armw UATsc _ _ _ => true
    | _ => false
  end.

(** Is the action a release write or stronger? *)

Definition is_release a :=
  match a with
    | Astore WATrel _ _ 
    | Astore WATsc _ _ 
    | Armw UATsc _ _ _
    | Armw UATacq_rel _ _ _ 
    | Armw UATcons_rel _ _ _
    | Armw UATrel _ _ _ => true
    | _ => false
  end.

(** Is the action an acquire read or stronger? *)
 
Definition is_acquire a :=
  match a with
    | Aload RATsc _ _ 
    | Aload RATacq _ _ 
    | Armw UATsc _ _ _
    | Armw UATacq_rel _ _ _
    | Armw UATacq _ _ _ => true
    | _ => false
  end.

(** Is the action a relaxed write or an RMW with a relaxed write ? *)

Definition is_rlx_write a := 
  match a with 
    | Astore WATrlx _ _ => true
    | Armw UATacq _ _ _ => true
    | Armw UATcons _ _ _ => true
    | Armw UATrlx _ _ _ => true
    | _ => false
  end.


(** Is the action a relaxed read or an RMW with a relaxed read ? *)

Definition is_rlx_read a := 
  match a with 
    | Aload RATrlx _ _  => true
    | Armw UATrel _ _ _ => true
    | Armw UATrlx _ _ _ => true
    | _ => false
  end.

(** Is the action a consume? *)

Definition is_consume a :=
  match a with
    | Aload RATcons _ _ => true
    | Armw UATcons _ _ _ => true
    | Armw UATcons_rel _ _ _ => true
    | _ => false
  end.

(** Is the action an atomic access? *)

Definition is_atomic a :=
  match a with
    | Aload RATna _ _ => false
    | Aload _ _ _ => true
    | Astore WATna _ _ => false
    | Astore _ _ _ => true
    | Armw _ _ _ _ => true
    | _ => false
  end.


(******************************************************************************)
(** ** Consistency axioms *)
(******************************************************************************)

Inductive model_type :=
  Mstd | MstdI | Mfixed | Mno_Wrlx.

Section Consistency.

Variable mt : model_type.
  (** Model type *)

Variable acts : list actid.  
  (** List of actions *)

Variable lab : actid -> action. 
  (** Labeling function *)

Variable sb : actid -> actid -> Prop. 
  (** "Sequence before" order *)

Variable dd : actid -> actid -> Prop.
  (** "Data dependence" order *)

Variable rf : actid -> option actid.  
  (** "Reads from" map *)

Variable mo : actid -> actid -> Prop. 
  (** Memory order *)

Variable mo_alt : actid -> actid -> Prop. 
  (** Alternative memory order that includes reads *)

Variable sc : actid -> actid -> Prop. 
  (** Sequential consistency order *)

Definition sameThread (a b : actid) :=
  clos_refl_trans _ sb a b.

Definition release_seq_elem (a b : actid) :=
  is_atomic (lab b) /\ (sameThread a b \/ is_rmw (lab b)).

Definition release_seq_elemI (a b : actid) :=
  is_atomic (lab b) /\ (sameThread a b \/ is_rmw (lab b) /\ rf b <> None).

Definition rmw_rf (a b : actid) :=
  is_rmw (lab b) /\ rf b = Some a.

Definition release_seq (a b : actid) :=
  match mt with
    | Mstd =>
        a = b \/ 
        mo a b /\ release_seq_elem a b 
        /\ forall c (MA: mo a c) (MB: mo c b), release_seq_elem a c
    | MstdI =>
        a = b \/ 
        mo a b /\ release_seq_elemI a b 
        /\ forall c (MA: mo a c) (MB: mo c b), release_seq_elemI a c
    | Mfixed =>
      exists c, clos_refl_trans _ sb a c /\ (a <> c -> is_atomic (lab c)) /\
                clos_refl_trans _ rmw_rf c b
    | Mno_Wrlx => a = b
  end.

Definition synchronizes_with (a b : actid) :=
  is_release (lab a) /\ is_acquire (lab b) /\
  exists c, release_seq a c /\ rf b = Some c.

Definition dep_reads_from (a b : actid) :=
  is_release (lab a) /\ is_consume (lab b) /\
  exists c, release_seq a c /\ rf b = Some c.

Definition dep_ordered_before (a b : actid) :=
  exists d,
    << DOBrf : dep_reads_from a d       >> /\
    << DOBdd : clos_refl_trans _ dd d b >>.

Definition syn_ordered_before (a b : actid) :=
  exists d,
    << SOBrf : synchronizes_with a d    >> /\
    << SOBsb : clos_refl_trans _ sb d b >>.

Definition ithb_one a b :=
  exists c,
    << OHBsb : clos_refl_trans _ sb a c >> /\
    << OHBsw : syn_ordered_before c b \/ dep_ordered_before c b >>.

Definition ithb := clos_trans _ ithb_one.

Definition happens_before a b := clos_trans _ sb a b \/ ithb a b.

Definition visible l a b := 
  happens_before a b /\
  (forall c (HB1: happens_before a c) (HB2: happens_before c b),
    ~ is_writeL (lab c) l).

Definition ExecutionFinite :=
  << CLOlab: forall a, lab a <> Askip -> In a acts >> /\
  << CLOsb : forall a b, sb a b -> In a acts /\ In b acts >>.

Definition AcyclicSB := acyclic sb.

Definition ConsistentDD :=
  (forall a b (DD: dd a b), clos_trans _ sb a b).

Definition ConsistentMO :=
  (forall a b (MO: mo a b), exists l, is_writeL (lab a) l /\ is_writeL (lab b) l)
  /\ transitive _ mo
  /\ (forall l, is_total (fun a => is_writeL (lab a) l) mo)
  /\ irreflexive mo
  /\ (forall l, restr_subset (fun a => is_writeL (lab a) l) happens_before mo).

Definition ConsistentMO_alt :=
  (forall a b (MO: mo_alt a b), exists l, is_accessL (lab a) l /\ is_accessL (lab b) l)
  /\ transitive _ mo_alt
  /\ (forall l, is_total (fun a => is_accessL (lab a) l) mo_alt)
  /\ irreflexive mo_alt
  /\ (forall l, restr_subset (fun a => is_accessL (lab a) l) happens_before mo_alt)
  /\ (forall a b (RF: rf b = Some a), immediate (fun a b => is_write (lab a) /\ mo_alt a b) a b).

Definition ConsistentMO_basic :=
  (forall a b (MO: mo a b), exists l, is_writeL (lab a) l /\ is_writeL (lab b) l)
  /\ transitive _ mo
  /\ (forall l, is_total (fun a => is_writeL (lab a) l) mo)
  /\ irreflexive mo.

Definition ConsistentSC :=
  (forall a b (MO: sc a b), is_sc (lab a) /\ is_sc (lab b))
  /\ transitive _ sc 
  /\ is_total (fun x => is_sc (lab x)) sc
  /\ restr_subset (fun x => is_sc (lab x)) happens_before sc
  /\ restr_subset (fun x => is_sc (lab x)) mo sc.

Definition ConsistentRF_basic :=
  forall a, 
    match rf a with
      | None => 
        forall b (HB: happens_before b a) 
               l (READ: is_readL (lab a) l) 
               (WRI: is_writeL (lab b) l), False
      | Some b =>
          exists l v, is_readLV (lab a) l v /\ is_writeLV (lab b) l v
    end.

Definition ConsistentRF_no_future :=
  forall a b (RF: rf b = Some a) (HB: happens_before b a), False.

Definition ConsistentRF_na :=
  forall a b (RF: rf a = Some b) (NA: is_na (lab a) \/ is_na (lab b)), 
    visible (loc (lab a)) b a.

Definition ConsistentRF_na2 :=
  forall a b (RF: rf a = Some b) (NA: is_na (lab a) \/ is_na (lab b)), 
    happens_before b a.

Definition SCrestriction := 
  forall a b (RF: rf b = Some a) (SC: is_sc (lab b)),
    immediate sc a b \/
    ~ is_sc (lab a) /\ (forall x (SC: immediate sc x b) (HB: happens_before a x), False).

Definition CoherentRR :=
  forall a b (HB: happens_before a b) 
         c (RFa: rf a = Some c) d (RFb: rf b = Some d) (MO: mo d c), 
    False.

Definition CoherentWR :=
  forall a b (HB: happens_before a b) 
         c (RF: rf b = Some c) (MO: mo c a),
    False.

Definition CoherentRW :=
  forall a b (HB: happens_before a b) 
         c (RF: rf a = Some c) (MO: mo b c),
    False.

Definition AtomicRMW := 
  forall a (RMW: is_rmw (lab a)) b (RF: rf a = Some b),
    immediate mo b a.

Definition ConsistentAlloc :=
  forall a la na (ALLOCa: lab a = Aalloc la na) 
         b lb nb (ALLOCb: lab b = Aalloc lb nb),
    a = b \/ la + na < lb \/ lb + nb < la.

Definition IrreflexiveITHB :=
  irreflexive ithb.

Definition AcyclicMedium :=
  acyclic (fun a b => happens_before a b \/ 
                      rf b = Some a /\ (is_na (lab b) \/ is_na (lab a))).

Definition AcyclicStrong :=
  acyclic (fun a b => sb a b \/ rf b = Some a).

Definition ConsistentC11 :=
  << FIN : ExecutionFinite >> /\
  << IRR : IrreflexiveITHB >> /\
  << Asb : AcyclicSB >> /\
  << Cdd : ConsistentDD >> /\
  << Cmo : ConsistentMO >> /\
  << Crf : ConsistentRF_basic >> /\
  << Crf': ConsistentRF_no_future >> /\
  << Cna : ConsistentRF_na >> /\
  << Csc : ConsistentSC >> /\
  << Csc': SCrestriction >> /\
  << Crr : CoherentRR >> /\
  << Cwr : CoherentWR >> /\
  << Crw : CoherentRW >> /\
  << Crmw: AtomicRMW >> /\
  << Ca  : ConsistentAlloc >>.

Definition ConsistentS :=
  << FIN : ExecutionFinite >> /\
  << ACYC : AcyclicStrong >> /\
  << Cdd : ConsistentDD >> /\
  << Cmo : ConsistentMO >> /\
  << Crf : ConsistentRF_basic >> /\
  << Crf': ConsistentRF_no_future >> /\
  << Cna : ConsistentRF_na2 >> /\
  << Csc : ConsistentSC >> /\
  << Csc': SCrestriction >> /\
  << Crr : CoherentRR >> /\
  << Cwr : CoherentWR >> /\
  << Crw : CoherentRW >> /\
  << Crmw: AtomicRMW >> /\
  << Ca  : ConsistentAlloc >>.

Definition ConsistentW :=
  << FIN : ExecutionFinite >> /\
  << ACYC : AcyclicStrong >> /\
  << Cdd : ConsistentDD >> /\
  << Cmo : ConsistentMO >> /\
  << Crf : ConsistentRF_basic >> /\
  << Crf': ConsistentRF_no_future >> /\
  << Csc : ConsistentSC >> /\
  << Csc': SCrestriction >> /\
  << Crr : CoherentRR >> /\
  << Cwr : CoherentWR >> /\
  << Crw : CoherentRW >> /\
  << Crmw: AtomicRMW >> /\
  << Ca  : ConsistentAlloc >>.

Definition ConsistentS_alt :=
  << FIN : ExecutionFinite >> /\
  << ACYC : AcyclicStrong >> /\
  << Cdd : ConsistentDD >> /\
  << Cmo : ConsistentMO_alt >> /\
  << Crf : ConsistentRF_basic >> /\
  << Cna : ConsistentRF_na2 >> /\
  << Csc : ConsistentSC >> /\
  << Csc': SCrestriction >> /\
  << Ca  : ConsistentAlloc >>.

Definition InitRMW_on_Mstd :=
  mt = Mstd -> forall a, is_rmw (lab a) -> rf a <> None.

(******************************************************************************)
(** ** Properties describing the absence of memory errors *)
(******************************************************************************)

Definition mem_safe := 
  forall a (Aa: is_access (lab a)),
  exists b l n, lab b = Aalloc l n /\ happens_before b a /\ l <= loc (lab a) <= l + n.

Definition initialized_reads := 
  forall a l (Aa: is_readL (lab a) l), rf a <> None.

Definition race_free := 
  forall a (Aa: is_access (lab a))
         b (Ab: is_access (lab b))
         (LOC: loc (lab a) = loc (lab b)) (NEQ: a <> b)
         (WRI: is_write (lab a) \/ is_write (lab b))
         (NA: is_na (lab a) \/ is_na (lab b)),
    happens_before a b \/ happens_before b a.

Definition rw_race_free := 
  forall a l (R: is_readL (lab a) l)
         b (Wb: is_writeL (lab b) l)
         (NA: is_na (lab a) \/ is_na (lab b)),
    happens_before a b \/ happens_before b a.

(******************************************************************************)
(** ** Lemmas about the memory model *)
(******************************************************************************)

Lemma sb_ithb_in_ithb a b c :
  clos_refl_trans _ sb a b ->
  ithb b c ->
  ithb a c.
Proof.
  induction 1; eauto; unfold ithb, ithb_one, NW.
  induction 1; desc; eauto using clos_trans.
  apply t_step; eauto using clos_refl_trans.
Qed.

Lemma sb_hb_in_hb a b c :
  clos_refl_trans _ sb a b ->
  happens_before b c ->
  happens_before a c.
Proof.
  unfold happens_before; ins; desf; eauto using sb_ithb_in_ithb, rt_t_trans.
Qed.

Lemma clos_trans_hb a b : 
  clos_trans actid happens_before a b <-> 
  clos_trans _ sb a b \/ exists c, ithb a c /\ clos_refl_trans _ sb c b.
Proof.
  split; ins; desf; vauto.
    unfold happens_before, ithb in *; apply clos_trans_t1n in H.
    induction H; ins; desf; eauto using rt_refl, t_trans, clos_trans_in_rt.
    right; eexists,; try edone.
    eapply sb_ithb_in_ithb; eauto using clos_trans_in_rt.
  induction H0 using clos_refl_trans_ind_left; vauto.
  eapply t_trans, t_step; vauto. 
Qed.

Lemma acyclicHB : irreflexive ithb -> acyclic sb -> acyclic happens_before.
Proof.
  unfold acyclic, irreflexive; ins.
  rewrite clos_trans_hb in *; desf; eauto using sb_ithb_in_ithb.
Qed.

Lemma acyclicHB1 : acyclic happens_before -> irreflexive ithb.
Proof.
  unfold acyclic, irreflexive, happens_before; ins; eauto using t_step.
Qed.

Lemma acyclicHB2 : acyclic happens_before -> irreflexive sb.
Proof.
  unfold acyclic, irreflexive, happens_before; ins; eauto using t_step.
Qed.

Lemma ConsistentNA1 
  (Crf : ConsistentRF_basic)
  (Cwr : CoherentWR) 
  (Cmo : ConsistentMO) : 
  ConsistentRF_na <-> ConsistentRF_na2.
Proof.
  unfold ConsistentRF_na, ConsistentRF_na2, visible.
  repeat split; ins; exploit H; eauto; clear NA; ins; desf. 
  intro W.
    exploit Cwr; try exact HB2; eauto. 
  red in Cmo; desf; eapply Cmo3; eauto.
  specialize (Crf a); rewrite RF in Crf; desf.
  by destruct (lab a), (lab b); ins; desf.
Qed.

Lemma wf_mo (FIN : ExecutionFinite) (Cmo : ConsistentMO_basic) : well_founded mo.
Proof.
  apply proj1 in FIN; unfold NW in *.
  constructor; ins. 
  assert (IN: In y acts).
    ins; apply FIN; apply (proj1 Cmo) in H; desf.
    by intro X; rewrite X in *.
  assert (PIN: forall a, mo a y -> In a acts).
    ins; apply FIN; apply (proj1 Cmo) in H0; desf.
    by intro X; rewrite X in *.
  clear FIN.
  red in Cmo; desf; clear Cmo1.
  remember (length acts) as n.
  clear H.
  revert IN PIN Heqn. 
  generalize acts as l.
  induction[y] n; ins; [by destruct l|]. 
  apply In_split in IN; desf.
  rewrite app_length in *; ins.
  rewrite NPeano.Nat.add_succ_r in *; desf.
  rewrite <- app_length in *; ins.

  constructor; ins; eapply IHn; try done; eauto.
    specialize (PIN _ H); rewrite in_app_iff in *; ins; desf; eauto.
    eby edestruct Cmo2.
  ins; exploit (PIN a0); eauto; rewrite !in_app_iff; ins; desf; eauto.
  by edestruct Cmo2; eauto.
Qed.


Lemma release_seq_weaken :
  forall 
    (FIN : ExecutionFinite) 
    (Cmo: ConsistentMO_basic) 
    (Crmw: AtomicRMW)
    (RMWI: InitRMW_on_Mstd) a b
    (RS : release_seq a b),
  exists c : actid,
     clos_refl_trans actid sb a c /\
     (c <> a -> is_atomic (lab c)) /\
     clos_refl_trans actid rmw_rf c b.
Proof.
  unfold release_seq, release_seq_elem, release_seq_elemI, sameThread, InitRMW_on_Mstd.
  ins; destruct mt; desf; 
     (try solve [eexists; repeat split; eauto using clos_refl_trans, clos_refl_trans_mon; 
                 ins]).
  {
    revert RS0 RS1 RS2.
    induction b using (well_founded_ind (wf_mo FIN Cmo)); ins.
    specialize (RMWI eq_refl _ RS2).
    assert (RMW:=RS2); eapply Crmw in RS2; destruct (rf b) as [c|] eqn: RF; ins.
    specialize (RS2 _ eq_refl); red in RS2; desf.
    destruct (eqP c a) as [|NEQ]; desf. 
      by eexists,; [|split; [|eapply rt_step, conj]]; eauto using rt_refl. 
    cut (exists d, clos_refl_trans _ sb a d /\ (d <> a -> is_atomic (lab d)) /\
                   clos_refl_trans _ rmw_rf d c).
      eby ins; desf; eexists,; [|eapply conj, rt_trans, rt_step, conj]. 
    destruct (proj1 Cmo _ _ RS2); desf.
    destruct (proj1 Cmo _ _ RS); desf.
    assert (x0 = x) by (destruct (lab b); ins; desf; desf); subst.
    eapply Cmo in NEQ; eauto; desf; [by exfalso; eauto|].
    exploit (RS1 c); ins; desf; eauto using rt_refl.
    eapply H; eauto; ins; desf. 
    by red in Cmo; desf; eapply RS1; eauto.
  }
  {
    revert RS0 RS1 RS2.
    induction b using (well_founded_ind (wf_mo FIN Cmo)); ins.
    assert (RMW:=RS2); eapply Crmw in RS2; destruct (rf b) as [c|] eqn: RF; ins.
    specialize (RS2 _ eq_refl); red in RS2; desf.
    destruct (eqP c a) as [|NEQ]; desf. 
      by eexists,; [|split; [|eapply rt_step, conj]]; eauto using rt_refl. 
    cut (exists d, clos_refl_trans _ sb a d /\ (d <> a -> is_atomic (lab d)) /\
                   clos_refl_trans _ rmw_rf d c).
      eby ins; desf; eexists,; [|eapply conj, rt_trans, rt_step, conj]. 
    destruct (proj1 Cmo _ _ RS2); desf.
    destruct (proj1 Cmo _ _ RS); desf.
    assert (x0 = x) by (destruct (lab b); ins; desf; desf); subst.
    eapply Cmo in NEQ; eauto; desf; [by exfalso; eauto|].
    exploit (RS1 c); ins; desf; eauto using rt_refl.
    eapply H; eauto; ins; desf. 
    by red in Cmo; desf; eapply RS1; eauto.
  }
Qed.

Lemma release_seq_weaken2 :
  forall 
    (FIN : ExecutionFinite) 
    (Cmo: ConsistentMO_basic) 
    (Crmw: AtomicRMW)
    (RMWI: InitRMW_on_Mstd) a b
    (RS : release_seq a b),
  clos_refl_trans actid (fun a b => sb a b \/ rf b = Some a) a b.
Proof.
  ins; exploit release_seq_weaken; eauto; unfold rmw_rf; intro; desf.
  eapply rt_trans; eapply clos_refl_trans_mon; eauto;
  instantiate; ins; desf; eauto.
Qed.

Lemma ithb_in_ct
    (FIN : ExecutionFinite) 
    (Cdd : ConsistentDD)
    (Cmo: ConsistentMO_basic) 
    (Crmw: AtomicRMW)
    (RMWS: InitRMW_on_Mstd) a b :
  ithb a b -> clos_trans actid (fun a b => sb a b \/ rf b = Some a) a b.
Proof.
  induction 1; eauto using clos_trans. 
  red in H; desc.
  eapply rt_t_trans; eauto using clos_refl_trans_mon; clear OHBsb. 
  desf; red in OHBsw; desf; eapply t_rt_trans with d; eauto using clos_refl_trans_mon.
    by red in SOBrf; ins; desf; eapply rt_t_trans, t_step; 
       try eapply release_seq_weaken2; eauto.
    by red in DOBrf; ins; desf; eapply rt_t_trans, t_step; 
       try eapply release_seq_weaken2; eauto.
  red in Cdd; desf; clear DOBrf.
  induction DOBdd; eauto using clos_refl_trans, clos_refl_trans_mon, clos_trans_in_rt.
Qed.

Lemma hb_in_ct
    (FIN : ExecutionFinite) 
    (Cdd : ConsistentDD)
    (Cmo: ConsistentMO_basic) 
    (Crmw: AtomicRMW)
    (RMWS: InitRMW_on_Mstd) a b :
  happens_before a b -> clos_trans actid (fun a b => sb a b \/ rf b = Some a) a b.
Proof.
  destruct 1; ins; eauto using ithb_in_ct.
  induction H; eauto using clos_trans. 
Qed.

Lemma release_seq_refl a : release_seq a a.
Proof. red; desf; vauto; exists a; repeat split; ins; apply rt_refl. Qed.

Lemma finiteRF : 
  forall (FIN: ExecutionFinite)
         (Crf: ConsistentRF_basic) a b,
    rf a = Some b -> In a acts /\ In b acts.
Proof.
  unfold ExecutionFinite; ins; desf.
  specialize (Crf a); desf; desf.
  generalize (CLOlab a); destruct (lab a); ins;
  generalize (CLOlab b); destruct (lab b); ins; intuition congruence.
Qed. 

Lemma race_free1 : race_free -> rw_race_free.
Proof.
  red; ins; apply H in NA; eauto; try intro; clarify;
  solve [try destruct (lab a); ins; destruct (lab b); ins; desf; vauto].
Qed.

Lemma ConsistentS_C11 :
  ConsistentS -> InitRMW_on_Mstd -> ConsistentC11.
Proof.
  unfold ConsistentS, ConsistentC11; ins; desf; unfold NW.
  rewrite ConsistentNA1; ins.
  intuition; repeat red; ins.
    by apply (ACYC x), ithb_in_ct; ins; red in Cmo; desf.
    by eauto using clos_trans_mon.
(* eapply ACYC, t_trans, t_step; eauto.
  red in HB; desf; eauto using clos_trans_mon.
  by eapply ithb_in_ct; ins; red in Cmo; desf. *)
Qed.

Lemma ConsistentC11_S :
  forall (NRW: forall a, ~ is_rlx_write (lab a))
         (NRR: forall a, ~ is_rlx_read (lab a)),
    ConsistentC11 -> ConsistentS.
Proof.
  unfold ConsistentS, ConsistentC11; ins; desf; unfold NW.
  rewrite ConsistentNA1 in *; ins.
  repeat (apply conj; try done); repeat red; ins.
  eapply acyclicHB with x; ins.
  revert H; generalize x at 1 3; induction 1; eauto using clos_trans; desf.
    by apply t_step, or_introl, t_step.
  specialize (Crf y); rewrite H in *; desf.
  specialize (NRW x); specialize (NRR y); specialize (Cna _ _ H).
  
  assert (X: is_release (lab x) \/ lab x = Astore WATna l v)
    by (destruct (lab x) as () eqn:?; ins; desf; vauto).
  desf; [|by apply t_step, Cna; right; rewrite X].
  apply t_step;
  destruct (lab y) as [] eqn: M; ins; destruct typ; ins; desf;
    try (by apply Cna; vauto); clear Cna;
  try solve [
  right; apply t_step; eexists,; eauto using rt_refl;
  left; eexists,; eauto using rt_refl; split; ins;
  rewrite M; eexists,,; eauto using release_seq_refl ];
  solve [
  right; apply t_step; eexists,; eauto using rt_refl;
  right; eexists,; eauto using rt_refl; split; ins;
  rewrite M; eexists,,; eauto using release_seq_refl ].
Qed. 

Lemma is_writeLV_writeL x l v : is_writeLV x l v -> is_writeL x l.
Proof. by destruct x; ins; desf. Qed.

Lemma is_writeL_write x l : is_writeL x l -> is_write x.
Proof. by destruct x. Qed.

Lemma is_only_readL_only_read x l : is_only_readL x l -> is_only_read x.
Proof. by destruct x. Qed.

Lemma is_writeL_accessL x l : is_writeL x l -> is_accessL x l.
Proof. by destruct x. Qed.

Lemma is_readLV_readL x l v : is_readLV x l v -> is_readL x l.
Proof. by destruct x; ins; desf. Qed.

Lemma is_readL_accessL x l : is_readL x l -> is_accessL x l.
Proof. by destruct x; ins; desf. Qed.

Lemma is_only_readL_readL x l : is_only_readL x l -> is_readL x l.
Proof. by destruct x; ins; desf. Qed.

Hint Resolve is_writeLV_writeL is_writeL_write is_writeL_accessL 
     is_readLV_readL is_readL_accessL is_only_readL_only_read is_only_readL_readL.

Lemma Consistent_altE :
  forall
    (Cmo : ConsistentMO_alt)
    (Crf : ConsistentRF_basic)
    (EQ : forall a b, mo a b <-> mo_alt a b /\ is_write (lab a) /\ is_write (lab b)),
  ConsistentMO /\ ConsistentRF_no_future /\
  CoherentRR /\ CoherentWR /\ CoherentRW /\ AtomicRMW.
Proof.
  unfold ConsistentMO_alt, immediate; ins; desf; repeat split; 
  try red; ins; repeat split; rewrite ?EQ in *; desf; eauto;
  repeat match goal with H : rf ?a = Some _ |- _ =>
    generalize (Crf a); rewrite H; intro; desf; apply Cmo4 in H end; 
  desf; eauto.

  by eapply Cmo in MO; desf; destruct (lab a); ins; destruct (lab b); vauto.
  by eapply Cmo1 in NEQ; desf; eauto 8.
  by split; try eapply Cmo3; eauto.
  by eapply Cmo3 with (l:=l) in HB; eauto.

  exploit (Cmo d c); eauto; intro; desf.
  assert (l1 = l) by (destruct (lab d); ins; desf); desf.
  assert (l0 = l) by (destruct (lab c); ins; desf); desf.
  by eapply Cmo3 with (l:=l) in HB; eauto.

  exploit (Cmo c a); eauto; intro; desf.
  assert (l0 = l) by (destruct (lab c); ins; desf); desf.
  by eapply Cmo3 with (l:=l) in HB; eauto.

  exploit (Cmo b c); eauto; intro; desf.
  assert (l0 = l) by (destruct (lab c); ins; desf); desf.
  by eapply Cmo3 with (l:=l) in HB; eauto.

  by repeat split; ins; destruct (lab a); ins.
Qed.

Lemma ConsistentSalt_S : 
  ConsistentS_alt -> 
  (forall a b, mo a b <-> mo_alt a b /\ is_write (lab a) /\ is_write (lab b)) ->
  ConsistentS.
Proof.
  unfold ConsistentS, ConsistentS_alt, NW; ins; desf.
  exploit Consistent_altE; eauto; tauto.
Qed.

Definition hb_or_order a b := 
  clos_trans _ happens_before a b \/ 
  ~ clos_trans _ happens_before a b /\ ~ clos_trans _ happens_before b a /\ a < b.

Require Import Classical. 

Lemma hb_or_order_tot :
  is_total (fun _ => True) hb_or_order.
Proof.
  red; unfold hb_or_order; ins.
  destruct (classic (clos_trans _ happens_before a b)); eauto.
  destruct (classic (clos_trans _ happens_before b a)); eauto.
  destruct (NPeano.Nat.compare_spec a b); desf; eauto.
Qed.

Definition mo_alt_def :=
  clos_trans _ (fun a b => 
     mo a b \/ rf b = Some a \/ (exists c, a <> b /\ rf a = Some c /\ mo c b)
     \/ (exists l, rf a = None /\ is_only_readL (lab a) l /\ is_writeL (lab b) l) 
     \/ (exists l, rf a = rf b /\
                     is_only_readL (lab a) l /\ is_only_readL (lab b) l
                     /\ hb_or_order a b)).

Ltac mo_des_contra H H' a := solve [clear - H H'; destruct a; ins; desf].

Ltac mo_clar Crf Cmo :=
  repeat match goal with 
    | H : ?P , H' : ?P |- _ => clear H'
    | H : is_only_read ?a, H' : is_write ?a |- _ => mo_des_contra H H' a
    | H : is_only_read ?a, H' : is_writeL ?a _ |- _ => mo_des_contra H H' a
    | H : is_only_readL ?a _, H' : is_writeL ?a _ |- _ => mo_des_contra H H' a
    | H : is_only_read ?a, H' : is_writeLV ?a _ _ |- _ => mo_des_contra H H' a
    | H : is_only_readL ?a _, H' : is_writeLV ?a _ _ |- _ => mo_des_contra H H' a
    | H : is_writeL ?a ?l1, H' : is_writeL ?a ?l2 |- _ =>
       assert (l2 = l1) by (destruct a; ins; desf); clear H'; desf
    | H : is_writeLV ?a ?l1 _, H' : is_writeL ?a ?l2 |- _ =>
       assert (l2 = l1) by (destruct a; ins; desf); clear H'; desf
    | H : is_readLV ?a ?l1 _, H' : is_writeL ?a ?l2 |- _ =>
       assert (l2 = l1 /\ is_rmw a) by (destruct a; ins; desf); revert H'; desf
    | H : is_only_readL ?a ?l1, H' : is_readLV ?a ?l2 _ |- _ =>
       assert (l2 = l1) by (destruct a; ins; desf); revert H'; desf
    | H : is_readLV ?a ?l1 ?v1, H' : is_writeLV ?a ?l2 ?v2 |- _ =>
       assert (l2 = l1 /\ v2 = v1 /\ is_rmw a) by (destruct a; ins; desf); revert H'; desf
    | H : is_writeLV ?a ?l1 ?v1, H' : is_writeLV ?a ?l2 ?v2 |- _ =>
       assert (l2 = l1 /\ v2 = v1) by (destruct a; ins; desf); clear H'; desf
    | H : rf ?a = Some _ |- _ =>
       let l := fresh "l" in
       let v := fresh "v" in
       generalize (Crf a); rewrite H; intros (l & v & ? & ?); revert H
    | H : mo _ _ |- _ => 
       let l := fresh "l" in
       destruct (Cmo _ _ H) as (l & ? & ?); revert H end; ins; eauto 7.

Lemma mo_alt_def_from_write 
  (Crf : ConsistentRF_basic)
  (Cmo : forall a b : actid,
        mo a b -> exists l : nat, is_writeL (lab a) l /\ is_writeL (lab b) l)
  (Cmo0: transitive _ mo)
  (Cmo1 : forall l : nat, is_total (fun a : actid => is_writeL (lab a) l) mo)
  (Cmo2: irreflexive mo)
  (Crmw: AtomicRMW)
 a b (MO : mo_alt_def a b) (W: is_write (lab a)) :
  mo a b \/ 
  is_only_read (lab b) /\ (rf b = Some a \/ exists c, rf b = Some c /\ mo a c).
Proof.
  ins; apply clos_trans_tn1 in MO; induction MO; desc; eauto.
    desf; try (by destruct (lab a)); eauto.
 
     destruct (classic (is_only_read (lab y))); eauto.
     by left; apply Crmw; ins; generalize (Crf y); 
        rewrite H; ins; desf; destruct (lab y); ins.
   destruct (eqP c a) as [|NEQ]; desf; eauto. 
   exploit Cmo; eauto; intro; desf.
   generalize (Crf a); rewrite H0; ins; desf.
   assert (l0 = l) by (destruct (lab c); ins; desf); desf.
   assert (is_rmw (lab a)) by (destruct (lab a); ins; desf).
   assert (is_writeL (lab a) l) by (destruct (lab a); ins; desf).
   eapply Cmo1 in NEQ; eauto; desf; eauto.
   eapply Cmo1 with (l:=l) in H; eauto; desf; eauto.
   by exfalso; apply Crmw, proj2 in H0; ins; desf; eauto. 

  clear MO; desf; mo_clar Crf Cmo; try rewrite H in *; eauto 8.
    destruct (classic (is_only_read (lab z))); eauto 7.
    left; apply Crmw, proj1 in H2; ins; eauto.
    by destruct (lab z); ins.
  left; eapply Cmo1 with (l:=l) in H; desf; eauto.
  by exfalso; eapply Crmw, proj2 in H7; ins; eauto.
Qed.


Lemma mo_alt_tot_helper : forall
  (Cmo : forall a b : actid,
        mo a b -> exists l : nat, is_writeL (lab a) l /\ is_writeL (lab b) l)
  (Cmo1 : forall l : nat, is_total (fun a : actid => is_writeL (lab a) l) mo)
  (Crf : ConsistentRF_basic) a b l
  (Wa : is_writeL (lab a) l)
  (Rb: is_only_readL (lab b) l),
  a = b \/ mo_alt_def a b \/ mo_alt_def b a.
Proof.
  ins; unfold mo_alt_def.
  destruct (eqP a b) as [|NEQ]; eauto; right.
  destruct (rf b) as [br|] eqn: RFb.
    mo_clar Crf Cmo.
    destruct (eqP br a) as [|NEQ']; desf; eauto using t_step.
    by eapply Cmo1 with (l:=l) in NEQ'; desf;
      [right; apply t_step|left; eapply t_trans; apply t_step |]; eauto 7.
  by right; apply t_step; do 3 right; eauto. 
Qed.

Lemma is_accessL_des x l : is_accessL x l -> is_writeL x l \/ is_only_readL x l.
Proof. destruct x; ins; vauto. Qed.

Lemma mo_alt_tot : forall
  (Cmo : forall a b : actid,
        mo a b -> exists l : nat, is_writeL (lab a) l /\ is_writeL (lab b) l)
  (Cmo1 : forall l : nat, is_total (fun a : actid => is_writeL (lab a) l) mo)
  (Crf : ConsistentRF_basic) l,
  is_total (fun a : actid => is_accessL (lab a) l) mo_alt_def.
Proof.
  red; ins; desf.
  apply is_accessL_des in IWa; apply is_accessL_des in IWb; desf.
    by unfold mo_alt_def; eapply Cmo1 in NEQ; desf; eauto using t_step.
    by eapply mo_alt_tot_helper in IWa; eauto; desf; eauto.
    by eapply mo_alt_tot_helper in IWb; eauto; desf; eauto.
  destruct (rf a) as [ar|] eqn: RFa.
    destruct (rf b) as [br|] eqn: RFb; mo_clar Crf Cmo.
    2: by right; eapply t_trans; apply t_step; eauto 10.
    destruct (eqP ar br) as [|NEQ']; desf.
      rewrite <- RFa in *.
      by eapply hb_or_order_tot in NEQ; desf; 
         [left|right]; apply t_step; do 4 right; eauto.
(*    destruct (eqP a br); desf. *)
    by eapply Cmo1 with (l:=l) in NEQ'; desf; eauto; 
       [left|right]; (eapply t_trans; apply t_step; [do 2 right; left|]; eauto);
       repeat eexists; eauto; intro; desf; mo_clar Crf Cmo. 
  destruct (rf b) as [br|] eqn: RFb; mo_clar Crf Cmo.
    by left; eapply t_trans; apply t_step; eauto 10.
 eapply hb_or_order_tot in NEQ; desf; [left|right]; apply t_step; 
 do 4 right; eauto 7; repeat eexists; eauto; congruence.
Qed.

Lemma mo_alt_def_range 
  (Cmo : forall a b : actid,
        mo a b -> exists l : nat, is_writeL (lab a) l /\ is_writeL (lab b) l)
  (Crf : ConsistentRF_basic) a b (MO : mo_alt_def a b) :
  exists l : nat, is_accessL (lab a) l /\ is_accessL (lab b) l.
Proof.
  induction MO; desc; eauto. 
  2: by assert (l = l0) by (destruct (lab y); ins; desf); desf; eauto.
  desf; mo_clar Crf Cmo.
Qed.

Lemma ConsistentS_Salt 
    (CONS: ConsistentS)
    (EQ: mo_alt = mo_alt_def)
    (RMWI: InitRMW_on_Mstd) :
  ConsistentS_alt.
Proof.
  unfold ConsistentS, ConsistentS_alt in *; unfold NW; desf; intuition.
  red in Cmo; desf; red; rewrite EQ; repeat split; try red; ins; desf; vauto.
  by apply mo_alt_def_range.
 
  eby eapply mo_alt_tot.

  exploit mo_alt_def_range; eauto; intro M; desf; mo_clar Crf Cmo.
  apply is_accessL_des in M; desf.
    by apply mo_alt_def_from_write in H; desf; mo_clar Crf Cmo.
clear l M.
  generalize H; revert H; generalize x at 1 4.
    induction 1; [|by unfold mo_alt_def in *; eauto using t_trans].
    ins; desf; mo_clar Crf Cmo.
    by apply mo_alt_def_from_write in H0; desf; mo_clar Crf Cmo.
    assert (N: mo_alt_def x x) by (eapply t_trans; [apply t_step|]; eauto).
    by apply mo_alt_def_from_write in N; desf; mo_clar Crf Cmo.

    by apply mo_alt_def_from_write in H0; desf; mo_clar Crf Cmo;
       eapply Crmw, proj2 in H7; eauto.
    by apply mo_alt_def_from_write in H0; desf; mo_clar Crf Cmo.
(*    cut (hb_or_order y x); [|clear H3].
      unfold hb_or_order in *; ins; desf; [|omega].
      by destruct (ACYC x); eapply t_trans; apply hb_in_ct; eauto. *)

    assert (~ hb_or_order x y). 
clear H3.

    apply clos_trans_t1n in H0.
    induction H0. 
      desf; mo_clar Crf Cmo.
admit.

    desf; mo_clar Crf Cmo; rewrite H in *.
      apply clos_t1n_trans, mo_alt_def_from_write in H3; desf; mo_clar Crf Cmo;
      exfalso; rewrite H in *; desf; eauto.

      apply clos_t1n_trans, mo_alt_def_from_write in H3; desf; mo_clar Crf Cmo;
      exfalso; rewrite H in *; desf; eauto.

    assert (l0 = l) by (destruct (lab x); ins; desf); desf.
      intro.
      exploit IHclos_trans_1n; eauto; try congruence.

clear - H6 H7.
unfold hb_or_order in *; ins; desf; eauto.




      apply clos_t1n_trans, mo_alt_def_from_write in H3; desf; mo_clar Crf Cmo.
      exfalso; rewrite H in *; desf; eauto.

    exfalso; rewrite H in *; desf; eauto.

admit.
omega.
intros [?|?].

admit.
    ins.

  assert (is_only_read (lab x) \/ lab x = Askip).

  red in H.

  assert (is_only_read (lab x) \/ lab x = Askip).
    (destruct (lab x)); ins; desf; eauto.
admit.
admit.

  apply is_accessL_des in IWa; apply is_accessL_des in IWb; desf.
    by eapply Cmo3 in REL; try apply t_step; eauto.
  destruct (rf a) as [ar|] eqn: RFa.
  2: by apply t_step; do 3 right; eauto.
  match goal with H : rf ?aa = Some _  |- _  =>
    let l0 := fresh "l" in 
    generalize (Crf aa); rewrite H; intros (l0 & ? & ? & ?); 
    assert (l0 = l) by (destruct (lab aa); ins; desf); desf
  end.

  destruct (eqP ar b) as [|NEQ']; desf; [by eapply Crf' in REL|].
  eapply Cmo1 in NEQ'; eauto; desf; eauto; [|by eapply Crw in NEQ'; eauto].
  destruct (eqP a b); desf.
    by destruct (ACYC b); apply hb_in_ct.
  by apply t_step; eauto 8.
  
  destruct (rf b) as [br|] eqn: RFb.
  match goal with H : rf ?aa = Some _  |- _  =>
    let l0 := fresh "l" in 
    generalize (Crf aa); rewrite H; intros (l0 & ? & ? & ?); 
    assert (l0 = l) by (destruct (lab aa); ins; desf); desf
  end.

  destruct (eqP br a) as [|NEQ']; desf; [by apply t_step; eauto|].
  eapply Cmo1 in NEQ'; eauto; desf; eauto; [by eapply Cwr in NEQ'; eauto|].
  by eapply t_trans; apply t_step; eauto 8.

  by exfalso; specialize (Crf b); rewrite RFb in *; eapply Crf; eauto. 

  destruct (rf a) as [ar|] eqn: RFa;
  destruct (rf b) as [br|] eqn: RFb;
  repeat match goal with H : rf ?aa = _  |- _  =>
    let l0 := fresh "l" in 
    generalize (Crf aa); rewrite H; try intros (l0 & ? & ? & ?); revert H; 
    try assert (l0 = l) by (destruct (lab aa); ins; desf); desf
  end; ins.

       destruct (eqP ar br) as [|MO]; [|eapply Cmo1 with (l:=l) in MO]; desf; eauto.
         by rewrite <- RFa in *; apply t_step; do 4 right; repeat eexists; eauto; vauto. 
       destruct (eqP a br); desf; [by apply t_step; eauto 8|].
       by eapply t_trans with br; apply t_step; eauto 8.
     by edestruct Crr; eauto.
admit.
  by apply t_trans with br; apply t_step; eauto 10.
  by rewrite <- RFa in *; apply t_step; do 4 right; repeat eexists; eauto; vauto.

  by specialize (Crf b); rewrite RF in *; desf; destruct (lab a); desf. 
  by apply t_step; eauto.

  apply mo_alt_def_from_write in R3; desf; 
    mo_clar Crf Cmo; try (by destruct (lab c)).
  apply mo_alt_def_from_write in R0; desf;
    mo_clar Crf Cmo. 
  by eapply Crmw, proj2 in RF; eauto.
Qed.


Lemma ConsistentS_W : ConsistentS -> ConsistentW.
Proof.
  unfold ConsistentS, ConsistentW, NW; tauto. 
Qed.

Lemma ConsistentC11_W :
  forall (NRW: forall a, ~ is_rlx_write (lab a))
         (NRR: forall a, ~ is_rlx_read (lab a)),
    ConsistentC11 -> ConsistentS.
Proof.
  auto using ConsistentS_W, ConsistentC11_S.
Qed.

Lemma ConsistentW_S :
  ConsistentW -> InitRMW_on_Mstd -> rw_race_free -> ConsistentS.
Proof.
  unfold ConsistentS, ConsistentW; ins; desf; unfold NW.
  repeat (apply conj; try done); red; ins.
  exploit finiteRF; eauto; intro; desc.
  specialize (Crf a); rewrite RF in *; desc.
  eapply H1 in NA; ins;
    try solve [clear NA; destruct (lab a); ins; destruct (lab b); ins; desf; vauto].
  desf; destruct (ACYC a); red in NA; ins; desf;
    eapply t_trans, t_step; eauto using clos_trans_mon.
  by eapply ithb_in_ct; ins; red in Cmo; desf.
Qed.

End Consistency.


Lemma hb_no_Wrlx :
  forall mt amap sb dd rf mo a b,
    happens_before Mno_Wrlx amap sb dd rf mo a b ->
    happens_before mt amap sb dd rf mo a b.
Proof.
  unfold happens_before, ithb, ithb_one, syn_ordered_before, dep_ordered_before,
  synchronizes_with, dep_reads_from; 
  ins; desf; eauto.
  right; induction H; desf; eauto using clos_trans; apply t_step; unfold NW; 
  eexists,; try edone; eauto 8 using release_seq_refl.
Qed. 

Lemma hb_trans_no_consumes : 
  forall amap (NC: forall a, ~ is_consume (amap a)) 
         mt sb dd rf mo a b c,
  happens_before mt amap sb dd rf mo a b -> 
  happens_before mt amap sb dd rf mo b c -> 
  happens_before mt amap sb dd rf mo a c.
Proof.
  unfold happens_before; ins; desf; 
  eauto using clos_trans, sb_ithb_in_ithb, clos_trans_in_rt; 
  right; vauto.
  unfold ithb, ithb_one, syn_ordered_before, dep_ordered_before,
    synchronizes_with, dep_reads_from, NW in *. 
  apply clos_trans_t1n in H; induction H; eauto; desf;
    try (eby edestruct NC); [|eapply t_trans; eauto];
    eapply t_step; eexists,; try edone; 
    left; eauto 10 using rt_trans, clos_trans_in_rt.
Qed.

Lemma hb_no_consumes : 
  forall amap (NC: forall a, ~ is_consume (amap a)) 
         mt sb dd rf mo a b,
  happens_before mt amap sb dd rf mo a b <-> 
  clos_trans _ (fun a b => sb a b \/ synchronizes_with mt amap sb rf mo a b) a b.
Proof.
  split; ins; [red in H|]; desf; eauto using clos_trans_mon.
    unfold happens_before, ithb, ithb_one, syn_ordered_before, dep_ordered_before, 
      dep_reads_from, NW in *.
    induction H; desf; eauto using clos_trans; try (eby edestruct NC).
    by eapply rt_t_trans, t_rt_trans; eauto using clos_refl_trans_mon, t_step.
  induction H; eauto using hb_trans_no_consumes; desf; vauto.
  by right; apply t_step; eexists,; [|left; eexists,]; eauto using rt_refl.
Qed.

Lemma hb_no_Wrlx2 :
  forall acts amap sb dd rf mo 
    (FIN : ExecutionFinite acts amap sb)
    (Crf: ConsistentRF_basic Mno_Wrlx amap sb dd rf mo) 
    (Cmo: ConsistentMO_basic amap mo) 
    (Crmw: AtomicRMW amap rf mo)
    (NRW: forall a, ~ is_rlx_write (amap a))
    (NRR: forall a, ~ is_rlx_read (amap a))
    (NC: forall a, ~ is_consume (amap a)) mt
    a b (HB: happens_before mt amap sb dd rf mo a b)
    (RMWS: InitRMW_on_Mstd mt amap rf),
    happens_before Mno_Wrlx amap sb dd rf mo a b.
Proof.
  ins; rewrite hb_no_consumes in HB; ins.
  induction HB; desf; vauto; eauto using hb_trans_no_consumes.
  red in H; desf.
  edestruct release_seq_weaken as (a & X & Y); eauto.
  eapply sb_hb_in_hb; eauto.
  generalize (Crf y); ins; desf; desf. 
  assert (Z: amap c = Astore WATna l v \/ is_release (amap c)).
    by (specialize (NRW c); destruct (amap c) as () eqn:?; ins; desf; eauto).
  desf.
    exfalso; clear H1; eapply clos_rt_rtn1 in Y0; induction Y0; eauto.
      by destruct (eqP a x); desf; rewrite Z in *; ins; desf; eauto.
    by red in H1; desf; rewrite Z in *; ins.
  clear H1 X.
  assert (happens_before Mno_Wrlx amap sb dd rf mo c y).
     by eapply hb_no_consumes, t_step; ins; right; repeat eexists; ins.
  clear Heq Z H2.
  eapply clos_rt_rt1n in Y0; induction Y0; eauto.
  red in H2; ins; desf; eapply hb_trans_no_consumes, IHY0; eauto.
  2: by destruct (amap y0).
  generalize (Crf y0); ins; desf; desf. 
  apply hb_no_consumes, t_step, or_intror; repeat eexists; ins. 
    specialize (NRW x0); destruct (amap x0) as () eqn:?; ins; desf; eauto.
    by eapply Y; intro; desf; rewrite Heqa in *. 
  by generalize (NRR y0), (NC y0); destruct (amap y0) as () eqn:?; ins; desf; eauto.
Qed. 

Lemma release_seq_elem1 amap sb rf a b : 
  release_seq_elemI amap sb rf a b ->
  release_seq_elem amap sb a b.
Proof.
  ins; red; red in H; tauto.
Qed.

Lemma release_seq_elem2 amap sb rf a b : 
  InitRMW_on_Mstd Mstd amap rf ->
  release_seq_elem amap sb a b ->
  release_seq_elemI amap sb rf a b.
Proof.
  ins; red; red in H; red in H0; desf; eauto. 
Qed.

Lemma hb_std_stdI :
  forall amap sb dd rf mo a b,
    happens_before MstdI amap sb dd rf mo a b ->
    happens_before Mstd amap sb dd rf mo a b.
Proof.
  unfold happens_before, ithb, ithb_one, syn_ordered_before, dep_ordered_before,
  synchronizes_with, dep_reads_from; ins; desf; eauto.
  right; induction H; desf; eauto using clos_trans; apply t_step; unfold NW;
  eexists,; try edone; eauto 14 using release_seq_elem1. 
Qed. 

Lemma hb_stdI_std :
  forall amap sb dd rf mo a b,
    happens_before Mstd amap sb dd rf mo a b ->
    InitRMW_on_Mstd Mstd amap rf ->
    happens_before MstdI amap sb dd rf mo a b.
Proof.
  unfold happens_before, ithb, ithb_one, syn_ordered_before, dep_ordered_before,
  synchronizes_with, dep_reads_from; ins; desf; eauto.
  right; induction H; desf; eauto using clos_trans; apply t_step; unfold NW;
  eexists,; try edone; eauto 14 using release_seq_elem2. 
Qed. 

Lemma hb_fixed :
  forall acts amap sb dd rf mo
    (FIN : ExecutionFinite acts amap sb) 
    (Cdd : ConsistentDD sb dd)
    (Cmo: ConsistentMO_basic amap mo) 
    (Crmw: AtomicRMW amap rf mo) mt
    (RMWS: InitRMW_on_Mstd mt amap rf) a b
    (HB : happens_before mt amap sb dd rf mo a b),
    happens_before Mfixed amap sb dd rf mo a b.
Proof.
  unfold happens_before, ithb; ins; desf; eauto; right. 
  unfold ithb_one, syn_ordered_before, dep_ordered_before,
  synchronizes_with, dep_reads_from in *; 
  ins; desf; eauto.
  induction HB; desc; eauto using clos_trans. 
  apply t_step; unfold NW; desf;
  eexists,; try edone; eauto 10 using rt_refl;
  [left|right]; do 6 (eexists; try edone);
  exploit release_seq_weaken; eauto; ins; desf; eauto.
Qed.


(** Mno_Wrlx is the weakest model. *)

Lemma ConsistentW_no_Wrlx :
  forall mt acts amap sb dd rf mo sc
    (CONS: ConsistentW mt acts amap sb dd rf mo sc)
    (*NRW: forall a, ~ is_rlx_write (amap a)*),
  ConsistentW Mno_Wrlx acts amap sb dd rf mo sc.
Proof.
  unfold ConsistentW; ins; unfold NW; desf; red in Cmo; red in Csc; desf;
  repeat (apply conj; try done); try red; ins; eauto using hb_no_Wrlx.
    by eapply Cmo3; eauto using hb_no_Wrlx.
    by specialize (Crf a); desf; ins; eauto using hb_no_Wrlx.
  by specialize (Csc' _ _ RF SC); desf; eauto using hb_no_Wrlx.
Qed. 

(** In the absence of relaxed and consume accesses, it is also
    the strongest model. *)

Lemma ConsistentS_no_Wrlx :
  forall mt acts amap sb dd rf mo sc
    (CONS: ConsistentW Mno_Wrlx acts amap sb dd rf mo sc)
    (RMWS: InitRMW_on_Mstd mt amap rf)
    (NRW: forall a, ~ is_rlx_write (amap a))
    (NRR: forall a, ~ is_rlx_read (amap a))
    (NC: forall a, ~ is_consume (amap a)),
  ConsistentW mt acts amap sb dd rf mo sc.
Proof.
  unfold ConsistentW; ins; unfold NW; desf; red in Cmo; red in Csc; desf.
  assert (ConsistentMO_basic amap mo) by done.
  repeat (apply conj; try done); try red; ins; eauto 3 using hb_no_Wrlx2.
    by eapply Cmo3; eauto using hb_no_Wrlx2.
    by generalize (Crf a); desf; ins; eauto using hb_no_Wrlx2.
  by specialize (Csc' _ _ RF SC); desf; eauto using hb_no_Wrlx2.
Qed. 


(** If there are no uninitialized RMWs, Mfixed is the strongest model. *)

Lemma ConsistentW_fixed :
  forall acts amap sb dd rf mo sc
    (CONS: ConsistentW Mfixed acts amap sb dd rf mo sc) mt
    (RMWI: InitRMW_on_Mstd mt amap rf),
  ConsistentW mt acts amap sb dd rf mo sc.
Proof.
  unfold ConsistentW; ins; unfold NW; desf; red in Cmo; red in Csc; desf.
  assert (ConsistentMO_basic amap mo) by done.
  repeat (apply conj; try done); try red; ins; eauto using hb_fixed.
    by eapply Cmo3; eauto using hb_fixed.
    by specialize (Crf a); desf; ins; eauto using hb_fixed. 
  by specialize (Csc' _ _ RF SC); desf; eauto using hb_fixed. 
Qed.

Corollary ConsistentW_fixed2 :
  forall acts amap sb dd rf mo sc
    (CONS: ConsistentW Mfixed acts amap sb dd rf mo sc) mt
    (RMWI: InitRMW_on_Mstd mt amap rf)
    (FREE: rw_race_free mt amap sb dd rf mo),
  ConsistentC11 mt acts amap sb dd rf mo sc.
Proof.
  auto using ConsistentS_C11, ConsistentW_S, ConsistentW_fixed. 
Qed.

(** If there are no uninitialized RMWs, Mstd and MstdI are equivalent. *)

Lemma ConsistentW_stdI :
  forall acts amap sb dd rf mo sc
    (RMWI: InitRMW_on_Mstd Mstd amap rf),
  ConsistentW Mstd acts amap sb dd rf mo sc <->
  ConsistentW MstdI acts amap sb dd rf mo sc.
Proof.
  unfold ConsistentW; split; ins; unfold NW; desf; red in Cmo; red in Csc; desf;
  repeat (apply conj; try done); red; ins; eauto using hb_std_stdI;
  eauto using hb_stdI_std;
  try solve [
    by eapply Cmo3; eauto using hb_std_stdI, hb_stdI_std
  | by specialize (Crf a); desf; eauto using hb_std_stdI, hb_stdI_std
  | by specialize (Csc' _ _ RF SC); desf; eauto using hb_std_stdI, hb_stdI_std ].
Qed.

Lemma ConsistentC11_stdI :
  forall acts amap sb dd rf mo sc
    (RMWI: InitRMW_on_Mstd Mstd amap rf),
  ConsistentC11 Mstd acts amap sb dd rf mo sc <->
  ConsistentC11 MstdI acts amap sb dd rf mo sc.
Proof.
  unfold ConsistentC11; split; ins; unfold NW; desf; red in Cmo; red in Csc; desf;
  rewrite ConsistentNA1 in *; ins;
  repeat (apply conj; try done); red; ins; eauto using hb_std_stdI;
  eauto using hb_stdI_std;
  try solve [
    by eapply Cmo3; eauto using hb_std_stdI, hb_stdI_std
  | by specialize (Crf a); desf; eauto using hb_std_stdI, hb_stdI_std
  | by specialize (Csc' _ _ RF SC); desf; eauto using hb_std_stdI, hb_stdI_std
  | by eapply acyclicHB1; repeat red; ins; eapply acyclicHB; 
       eauto using clos_trans_mon, hb_std_stdI, hb_stdI_std ].
Qed.

(*
Definition act_similar act act' :=
  act = act' 
  \/ exists l l' n, act = Aalloc l n /\ act' = Aalloc l' n
  \/ exists t l v v', act = Aload t l v /\ act' = Aload t l v'
  \/ exists t l f v v', act = Armw t l f v /\ act' = Armw t l f v'.
*)


(*
Require Import Classical.

Lemma ConsistentW_read_extend : 
  forall mt acts amap sb dd rf mo sc
    (CONS: ConsistentW mt acts amap sb dd rf mo sc)
    b (NIN: ~ In b acts) sb_pres (SB1: forall x, In x sb_pres -> In x acts) 
    dd_pres (DD1: forall x, In x dd_pres -> In x acts)
    (DDinSB: forall x, In x dd_pres -> 
             exists y, In y sb_pres /\ clos_refl_trans _ sb x y)
    sb' (SB: forall x y, sb' x y <-> sb x y \/ In x sb_pres /\ y = b)
    dd' (DD: forall x y, dd' x y <-> dd x y \/ In x dd_pres /\ y = b) act,
  exists act', act_similar act act' /\ exists rf' mo' sc',
    ConsistentW mt (b :: acts) 
                (fun c => if c == b then act' else amap c) 
                sb' dd' rf' mo' sc'.
Proof.
  unfold ConsistentW; ins; unfold NW; desf; red in Cmo; red in Csc; desf;
  repeat (apply conj; try done).
  assert (FIN': forall act, 
            ExecutionFinite (b :: acts)
                            (fun c => if c == b then act else amap c) sb').
     by split; unfold NW; ins; desf; vauto; try rewrite SB in *; desf; 
        try eapply FIN in H; desf; eauto.
  assert (ACYC': forall rf', 
                   rf' b <> Some b -> 
                   (forall x, x <> b -> rf' x = rf x) ->
                   AcyclicStrong sb' rf').
    repeat red; ins; apply (ACYC x).
(*    destruct (eqP x b); desf.
      exfalso; rewrite t_step_rt in H1; desf.
        by rewrite SB in *; desf; eauto; eapply FIN in H1; desf.
      destruct (eqP z b); desf.
      by rewrite H0 in H1; ins; eapply finiteRF in H1; eauto; desf. 
    revert H1; generalize x at 2 4; ins.
      apply clos_trans_tn1 in H1; induction H1; desf.
        rewrite SB in *; desf; eauto; eapply FIN in H1; desf.
 *)
    admit.
  assert (Cdd' : ConsistentDD sb' dd').
    red; ins; rewrite DD in *; desf. 
      by eapply clos_trans_mon; eauto; ins; rewrite SB; eauto.
    exploit DDinSB; eauto; ins; desf.
    eapply rt_t_trans, t_step, SB; eauto.
    by eapply clos_refl_trans_mon; eauto; ins; rewrite SB; eauto.

  destruct (classic (exists l v, is_readLV act l v)); desf.
admit.
  destruct (classic (exists l, is_writeL act l)); desf.
admit.
  exists act; split; vauto.
  exists rf, mo, sc.
   repeat (apply conj; eauto). 
     by eapply ACYC'; ins; intro; eauto using t_step.
   ins. 
     eapply Cmo in MO; desf; desf; eauto;
     by assert (M: amap b <> Askip) by (intro X; rewrite X in *; ins);
        eapply FIN in M.
   red; ins; desf; desf; solve [exfalso; eauto | eapply Cmo1; eauto].
   red; ins; desf; desf; try solve [exfalso; eauto | eapply Cmo3; eauto].
     eapply Cmo3; eauto.
     admit.
admit.

   red.

destruct H0; 

 eapply Cmo1; eauto.

desf; desf.

   split; eauto.
split.
  red; repeat  red in Cmo; ins.
   


split.
red.


admit.

  eapply ACYC; eauto.
      
 induction 1; eauto using .
    rewrite SB in *; desf.

    exfalso. 
SearchAbout clos_trans.
inv H1; eauto. induction H1; eauto.

    generalize (eq_refl x); revert H1; generalize x at 1 3; induction 1. 
ins.
    ins; clarify; eauto using t_trans.
admit.
    rewrite SB in *; destruct (eqP x b); desf; eauto using t_step;
      try (by apply FIN in H1; desf);
      try destruct (eqP y b) as [|N]; desf;
      try (by exfalso; eauto);
      try (rewrite (H0 _ N) in * );
      try (by eapply finiteRF in H1; eauto; desf). 



    
admit.
     split; unfold NW; ins; desf; vauto; try rewrite SB in *; desf; 
        try eapply FIN in H; desf; eauto.

 


 desf.

 
   repeat (apply conj; try done). 

split; 

mo.

try red; ins; eauto using hb_no_Wrlx.
    by eapply Cmo3; eauto using hb_no_Wrlx.
    by specialize (Crf a); desf; ins; eauto using hb_no_Wrlx.
  by specialize (Csc' _ _ RF SC); desf; eauto using hb_no_Wrlx.
Qed. 
*)

