(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting Program extralib MMergesort. 
Require Import c11 exps.

Set Implicit Arguments.
Obligation Tactic := intros; desc; try done.

(** Generic pointwise lifting of a partial binary operator to a function,
    where the lifted operator is defined on two functions if and only if it
    is defined on all elements of those functions. *)

Lemma lift_plus_helper A B (p : A -> option B) (PF : forall x, p x <> None) : 
  { f | forall x, p x = Some (f x) }.
Proof.
  apply constructive_definite_description.
  rewrite <- unique_existence; split.
    apply choice with (R := fun x y => p x = Some y).
    by intro x; specialize (PF x); destruct (p x); vauto.
  clear; red; intros; apply functional_extensionality; intro a.
  specialize (H a); specialize (H0 a); congruence.
Qed.

Definition lift_plus A B p (f g : A -> B) : option (A -> B) := 
  match excluded_middle_informative (forall x, p (f x) (g x) <> None) with
    | left PF => Some (proj1_sig (lift_plus_helper (fun x => p (f x) (g x)) PF))
    | _ => None
  end.

Lemma lift_plus_someD A B p (f g : A -> B) h :
  lift_plus p f g = Some h <-> forall x, p (f x) (g x) = Some (h x). 
Proof.
  unfold lift_plus, proj1_sig; ins; desf; split; ins; desf. 
    f_equal; apply functional_extensionality; intro a; clear Heq.
    specialize (H a); specialize (e a); congruence.
  destruct n; congruence.
Qed.

Lemma lift_plus_noneD A B p (f g : A -> B) :
  lift_plus p f g = None <-> exists x, p (f x) (g x) = None. 
Proof.
  unfold lift_plus, proj1_sig; ins; desf; split; ins; desf. 
    by clear Heq; specialize (n x0); desf. 
  apply NNPP; intro M; destruct n; red; ins; eauto. 
Qed.

Lemma lift_plus_emp_l A B p (f : A -> B) e (E: forall x, p e x = Some x) : 
  lift_plus p (fun _ => e) f = Some f.
Proof. by rewrite lift_plus_someD; eauto. Qed.

Lemma lift_plusC B p (C: forall x y, p x y = p y x) A (f g : A -> B) : 
  lift_plus p f g = lift_plus p g f.
Proof. 
  unfold lift_plus, proj1_sig; desf; clear Heq; try clear Heq0.
    by f_equal; extensionality a; generalize (e a), (e0 a); rewrite C; 
       clear; congruence.
    by destruct n0; intros; rewrite C.
    by destruct n; intros; rewrite C.
Qed.

Lemma lift_plusA B p
  (ASSOC: forall x y xy (P1: p x y = Some xy) z xyz (P2: p xy z = Some xyz),
          exists yz, p y z = Some yz /\ p x yz = Some xyz)
  A (x y xy : A -> B) (P1: lift_plus p x y = Some xy) z xyz 
  (P2: lift_plus p xy z = Some xyz) :
    exists yz, lift_plus p y z = Some yz /\ lift_plus p x yz = Some xyz.
Proof.
  unfold lift_plus, proj1_sig in *; desf; clear Heq Heq0; try clear Heq1.
  eexists,; try done; desf; try clear Heq.
  f_equal; extensionality a.
  generalize (e a), (e0 a), (e1 a), (e2 a); clear e e0 e1 e2; ins.
  exploit (ASSOC (x a) (y a)); eauto; ins; desf.

  destruct n2; intros a ?.
  generalize (e a), (e0 a), (e1 a); clear e e0 e1; ins.
  exploit (ASSOC (x a) (y a)); eauto; ins; desf.

  destruct n; intros a ?.
  generalize (e a), (e0 a); clear e e0; ins.
  exploit (ASSOC (x a) (y a)); eauto; ins; desf.
Qed.

Lemma lift_plusA2 B p
  (ASSOC: forall x y xy (P1: p x y = Some xy) z (P2: p xy z = None),
    p y z = None \/ exists yz, p y z = Some yz /\ p x yz = None)
  A (x y xy : A -> B) (P1: lift_plus p x y = Some xy) z
  (P2: lift_plus p xy z = None) :
   lift_plus p y z = None \/ 
   exists yz, lift_plus p y z = Some yz /\ lift_plus p x yz = None.
Proof.
  unfold lift_plus, proj1_sig in *; desf; vauto.
  right; clear Heq Heq0; try clear Heq1.
  eexists,; try done; desf; try clear Heq.
  destruct n0; ins; intro.
  eapply ASSOC in H; eauto; desf; congruence. 
Qed.

Lemma lift_plus_some_inv A B p (a b c : A -> B) (EQ : lift_plus p a b = Some c) l :
  p (a l) (b l) = Some (c l).
Proof.
  unfold lift_plus in *; desf; destruct lift_plus_helper; ins.
Qed.

Lemma lift_plus_none_inv A B p (a b : A -> B) (EQ : lift_plus p a b = None) :
  exists l, p (a l) (b l) = None.
Proof.
  unfold lift_plus in *; desf; apply NNPP; intro; destruct n; red; eauto.
Qed.

Ltac inv_lplus a := 
  repeat match goal with
           | H : lift_plus _ _ _ = Some _ |- _ => 
             apply lift_plus_some_inv with (l:=a) in H
           | H : Some _ = lift_plus _ _ _ |- _ => 
             symmetry in H; apply lift_plus_some_inv with (l:=a) in H
         end.

Ltac inv_lplus2 a := 
  repeat match goal with
           | H : lift_plus _ _ _ = None |- _ => 
             apply lift_plus_none_inv in H; destruct H as [a H]
           | H : None = lift_plus _ _ _ |- _ => 
             symmetry in H; apply lift_plus_none_inv in H; destruct H as [a H]
           | H : lift_plus _ _ _ = Some _ |- _ => 
             apply lift_plus_some_inv with (l:=a) in H
           | H : Some _ = lift_plus _ _ _ |- _ => 
             symmetry in H; apply lift_plus_some_inv with (l:=a) in H
         end.


(******************************************************************************)
(******************************************************************************)
(** * 2.1. Semantic structures *)
(******************************************************************************)
(******************************************************************************)

(******************************************************************************)
(** ** 2.1.1. Parameters *)
(******************************************************************************)

(** Assume a ghost resource type: a partial commutative monoid. *)

Parameter elem_gres : Type.
Parameter elem_gres_emp : elem_gres. 
Parameter elem_gres_plus : elem_gres -> elem_gres -> option elem_gres.
Parameter elem_gres_strip : elem_gres -> elem_gres.

Axiom elem_gres_plusC : 
  forall x y, elem_gres_plus x y = elem_gres_plus y x.

Axiom elem_gres_plusI : 
  forall x, elem_gres_plus elem_gres_emp x = Some x.

Axiom elem_gres_plusA : 
  forall x y xy (P1: elem_gres_plus x y = Some xy) z xyz 
                (P2: elem_gres_plus xy z = Some xyz),
    exists yz, elem_gres_plus x yz = Some xyz /\ 
               elem_gres_plus y z = Some yz. 

Axiom elem_gres_plusA2 : 
  forall x y xy (P1: elem_gres_plus x y = Some xy) z 
                (P2: elem_gres_plus xy z = None),
    elem_gres_plus y z = None
    \/ exists yz, elem_gres_plus x yz = None /\ 
                  elem_gres_plus y z = Some yz. 

Axiom elem_gres_eq_empD : 
  forall g g', elem_gres_plus g g' = Some elem_gres_emp -> 
               g = elem_gres_emp.

Axiom elem_gres_strip_emp :
  elem_gres_strip elem_gres_emp = elem_gres_emp.

Axiom elem_gres_strip_plus :
  forall x y xy, elem_gres_plus x y = Some xy ->
    elem_gres_plus (elem_gres_strip x) (elem_gres_strip y) = Some (elem_gres_strip xy).

Axiom elem_gres_plus_strip_idemp :
  forall x, elem_gres_plus x (elem_gres_strip x) = Some x. 

Axiom elem_gres_strip_max :
  forall x y xy z, 
    elem_gres_plus x y = Some xy ->
    elem_gres_plus xy z = Some x ->
    elem_gres_strip y = y.

Axiom elem_gres_pcancel :
  forall g g1 g2 gg,
    elem_gres_plus g g1 = Some gg ->
    elem_gres_plus g g2 = Some gg -> 
    elem_gres_plus (elem_gres_strip g) g1 = 
    elem_gres_plus (elem_gres_strip g) g2.

Lemma elem_gres_strip_idemp :
  forall x, elem_gres_strip (elem_gres_strip x) = elem_gres_strip x.
Proof.
  ins; eapply elem_gres_strip_max; [by apply elem_gres_plus_strip_idemp|].
  by rewrite elem_gres_plusC; apply elem_gres_plusI.
Qed.

Inductive state    := mkState (n : nat).
Inductive escrowTy := mkEscrowTy (n : nat).
Inductive protTy   := mkProtTy (n : nat).

(** Decidable equality on escrow types *)

Definition eq_escrow (x y: escrowTy) :=
   match x, y with
     | mkEscrowTy n1, mkEscrowTy n2 => n1 == n2
   end.

Lemma eq_escrowP: forall x y, reflect (x = y) (eq_escrow x y).
Proof.
  induction[] x [y]; ins; case eqP; constructor; congruence.
Qed.

Canonical Structure escrowTy_eqMixin := EqMixin eq_escrowP.
Canonical Structure escrowTy_eqType := Eval hnf in EqType escrowTy escrowTy_eqMixin.

(** Assume a partial order per protocol type. *)

Parameter prot_trans : protTy -> state -> state -> Prop. 
Axiom prot_trans_ord : forall tau, order _ (prot_trans tau).

Lemma prot_trans_trans : forall tau x y z,
  prot_trans tau x y -> prot_trans tau y z -> prot_trans tau x z.
Proof.
  eby ins; eapply (ord_trans _ _ (prot_trans_ord _)).
Qed.

Lemma prot_trans_refl : forall tau x, prot_trans tau x x.
Proof.
  eby ins; eapply (ord_refl _ _ (prot_trans_ord _)).
Qed.

Definition dec_prot_trans tau x y :=
  if excluded_middle_informative (prot_trans tau x y) then true else false.

(******************************************************************************)
(** ** 2.1.2. Domains *)
(******************************************************************************)

(** Traces *)

Record slist (tau : protTy) := 
  { SL_list : list state ;
    SL_nemp : SL_list <> nil ;
    SL_ndup : NoDup SL_list ;
    SL_sort : StronglySorted (prot_trans tau) SL_list }.

Lemma Sorted_prot tau l : 
  (forall x y, In x l -> In y l -> prot_trans tau x y \/ prot_trans tau y x) ->
  StronglySorted (prot_trans tau) (sort (dec_prot_trans tau) l).
Proof.
  ins; exploit (sort_sorted (dec_prot_trans tau) l).
    by unfold dec_prot_trans; ins; desf; auto; 
       exfalso; specialize (H _ _ H0 H1); desf.
    by unfold dec_prot_trans; red; ins; desf; destruct n; 
       eauto using prot_trans_trans.
  clear; induction 1; econstructor; eauto.
  rewrite Forall_forall in *; ins; eapply H in H0; 
  unfold dec_prot_trans in *; desf.
Qed.

Program Definition mk_slist (tau : protTy) (l : list state) : option (slist tau) :=
  if excluded_middle_informative 
       (l <> nil /\ forall x y, In x l -> In y l -> 
                    prot_trans tau x y \/ prot_trans tau y x)
  then Some {| SL_list := undup (sort (dec_prot_trans tau) l) ;
               SL_ndup := nodup_undup _ |}
  else None.
Next Obligation. 
  apply undup_nonnil.
  intro X; destruct H; eapply Permutation_nil; rewrite <- X; apply sort_perm.
Qed.
Next Obligation. by apply Sorted_undup, Sorted_prot. Qed.

Program Definition mk_slist_singl tau s : slist tau :=
  {| SL_list := s :: nil |}.
Next Obligation. vauto. Qed.

Program Definition get_state tau (l : slist tau) : state :=
  match rev (SL_list l) with
    | nil => !
    | x :: _ => x
  end.
Next Obligation.
  destruct l as [[] ? ?]; ins.
  apply eq_sym, app_eq_nil in Heq_anonymous; desf.
Qed.

Lemma slist_eq tau (l l' : slist tau) :
  SL_list l = SL_list l' -> l = l'.
Proof.
  destruct l, l'; ins; desf; f_equal; apply proof_irrelevance.
Qed.

Lemma slist_eq2 tau (l l' : slist tau) :
  (forall x, In x (SL_list l) <-> In x (SL_list l')) ->
  l = l'.
Proof.
  intros; apply slist_eq; destruct l, l'; ins.
  eapply sorted_perm_eq; 
    eauto using NoDup_Permutation; try apply prot_trans_ord.
Qed.

Lemma sort_undup A (p : A -> A -> bool) 
    (T: transitive _ p) (AS: antisymmetric _ p) 
    (l : list A) 
    (TOTAL: forall x y, In x l -> In y l -> p x y = true \/ p y x = true) :
  sort p (undup l) = undup (sort p l).
Proof.
  eapply sorted_perm_eq; eauto using sort_sorted, Sorted_undup.
    by eapply NoDup_Permutation; eauto; ins; 
       rewrite In_sort, !in_undup_iff, In_sort.
  apply sort_sorted; ins; rewrite in_undup_iff in *; eauto.
Qed.

Lemma undup_sort_undup A p (l : list A) :
  undup (sort p (undup l)) = sort p (undup l).
Proof. apply undup_nodup; eauto. Qed.

Lemma mk_slist_eq tau l l' :
  (forall x, In x l <-> In x l') ->
  mk_slist tau l = mk_slist tau l'.
Proof.
  intros; unfold mk_slist; desf; f_equal.
    by apply slist_eq2; ins; rewrite !in_undup_iff, !In_sort.

  * exfalso; destruct n; split; try red; ins; desf.
      destruct l as [|aa ?]; ins; specialize (H aa); tauto. 
      by apply a0; apply H.

  * exfalso; destruct n; split; try red; ins; desf.
      destruct l' as [|aa ?]; ins; specialize (H aa); tauto. 
      by apply a0; apply H.
Qed.

Lemma In_mk_slist tau x l S :
  mk_slist tau l = Some S -> In x l -> In x (SL_list S).
Proof.
  by unfold mk_slist; ins; desf; ins; apply in_undup_iff, In_sort.
Qed.

Lemma mk_slist_cons_in tau s S S' :
  mk_slist tau (s :: SL_list S) = Some S' ->
  In s (SL_list S) ->
  S = S'.
Proof.
  unfold mk_slist in *; ins; desf; apply slist_eq; ins.
  destruct S; ins; desf.
  apply sorted_perm_eq with (cmp := prot_trans tau); try apply prot_trans_ord;
    eauto using Sorted_prot, Sorted_undup.
  apply NoDup_Permutation; ins; rewrite in_undup_iff, In_sort; split; ins; desf; vauto.
Qed.

Lemma get_state_in tau (S: slist tau) : In (get_state S) (SL_list S).
Proof.
  unfold get_state; destruct S; des_eqrefl; ins.
    by exfalso; induction SL_list0 using rev_ind; ins; rewrite rev_app_distr in *.
  apply in_rev; rewrite <- EQ; vauto.
Qed.

Lemma get_state_lt tau (S: slist tau) s (IN: In s (SL_list S)) :
  prot_trans tau s (get_state S).
Proof.
  unfold get_state; destruct S; des_eqrefl; ins.
    by exfalso; induction SL_list0 using rev_ind; ins; rewrite rev_app_distr in *.
  apply (f_equal (fun f => rev f)) in EQ; ins; rewrite rev_involutive in *; subst.
  rewrite in_app_iff in *; ins; desf; eauto using prot_trans_refl.
  eapply Sorted_app_end_inv; eauto.
Qed.



(** Protocols *)

Inductive protocol := 
  | p_bot
  | p_uninit
  | p_na (v : nat)
  | p_at (tau : protTy) (s : slist tau).

Arguments p_at : clear implicits. 

(** Ghost resources *)

Structure ghost_resource := 
  mk_gres { 
      gres_fun :> nat -> elem_gres ;
      gres_finite : 
        exists max, forall n', max < n' -> gres_fun n' = elem_gres_emp
    }.

(** Resources *)

Inductive resource := 
  | Rundef 
  | Rdef (pmap: nat -> protocol)
         (gres: ghost_resource) 
         (escr: escrowTy -> bool).

(******************************************************************************)
(** ** 2.1.3. Resource composition *)
(******************************************************************************)

(** Protocol composition *)

Definition prot_plus (pi1 pi2 : protocol) :=
  match pi1, pi2 with
    | p_bot, _ => Some pi2
    | _, p_bot => Some pi1
    | p_at tau1 s1, p_at tau2 s2 =>
      if excluded_middle_informative (tau1 = tau2) then
        match mk_slist tau1 (SL_list s1 ++ SL_list s2) with
          | Some s => Some (p_at tau1 s)
          | None => None
        end
      else None
    | _, _ => None
  end. 

(** Empty ghost resource *)

Program Definition gres_emp :=
  {| gres_fun := fun _ => elem_gres_emp |}.
Next Obligation. by exists 0. Qed.

(** Singleton ghost resource *)

Program Definition gres_one n g :=
  {| gres_fun := fun n' => if n==n' then g else elem_gres_emp |}.
Next Obligation. 
  exists (S n); ins; desf; desf; exfalso; omega.
Qed.

(** Ghost resource composition *)

Program Definition gres_plus (g1 g2 : ghost_resource) :=
  match lift_plus elem_gres_plus g1 g2 with
    | Some g => Some {| gres_fun := g |}
    | None => None
  end.
Next Obligation.
  destruct g1, g2; desf.
  exists (max + max0); ins; inv_lplus n'. 
  rewrite gres_finite0, gres_finite1, elem_gres_plusI in *; desf; omega.
Qed.

(** Ghost resource striping *)

Program Definition gres_strip (g : ghost_resource) :=
  {| gres_fun := fun l => elem_gres_strip (g l) |}.
Next Obligation.
  destruct g; desf; exists max; ins; inv_lplus n'. 
  by rewrite gres_finite0, elem_gres_strip_emp.
Qed.

(** Empty resource *)

Definition res_emp : resource :=
  Rdef (fun _ => p_bot) gres_emp (fun _ => false).

(** Resource composition *)

Definition res_plus (x y : resource) : resource := 
  match x, y with
    | Rundef, _ => Rundef
    | _, Rundef => Rundef 
    | Rdef pmap1 gres1 escr1, Rdef pmap2 gres2 escr2 =>
        match lift_plus prot_plus pmap1 pmap2 with
          | None => Rundef 
          | Some pmap =>
            match gres_plus gres1 gres2 with 
              | None => Rundef 
              | Some gres => 
                Rdef pmap gres (fun sigma => escr1 sigma || escr2 sigma)
            end
        end
  end.

(** Multiple resource composition *)

Fixpoint rsum rl :=
  match rl with
    | nil => res_emp
    | a :: rl => res_plus a (rsum rl)
  end.

(** Read location from resource *)

Definition res_get r loc := 
  match r with
    | Rundef => p_bot 
    | Rdef pmap _ _ => pmap loc
  end.

(** Resource update *)

Definition res_upd r loc pi :=
  match r with
    | Rundef => Rundef
    | Rdef pmap gres escr =>
      Rdef (fun l => if l == loc then pi else pmap l) gres escr
  end.

(** Resource multiple update *)

Definition res_mupd r loc n pi :=
  match r with
    | Rundef => Rundef
    | Rdef pmap gres escr => 
      Rdef (fun l => if excluded_middle_informative (loc <= l <= loc + n)
                     then pi else pmap l)
           gres escr
  end.

(** Read ghost location from resource *)

Definition res_gget r l :=
  match r with
    | Rundef => elem_gres_emp
    | Rdef p g e => g l
  end.

(** Read escrow from resource *)

Definition res_escr r sigma :=
  match r with 
      Rundef => False
    | Rdef p g e => e sigma
  end.

(** Trace comparison *)

Definition slist_lt tau (S: slist tau) tau' (S': slist tau') :=
  forall s (SIN: In s (SL_list S)),
  exists s', In s' (SL_list S') /\ prot_trans tau s s'.

(** Protocol transition is allowed *)

Definition prot_trans_ok (pi1 pi2: protocol) := 
  match pi1, pi2 with
    | p_at tau1 sl1 , p_at tau2 sl2 => 
      tau1 = tau2 /\ 
      slist_lt sl1 sl2 
    | _, _ => False
  end.

(** ACI properties of protocol composition *)

Lemma prot_plus_emp_l x : prot_plus p_bot x = Some x.
Proof. done. Qed.

Lemma prot_plus_emp_r x : prot_plus x p_bot = Some x.
Proof. by destruct x. Qed.

Lemma prot_plusC :
  forall x y, prot_plus x y = prot_plus y x.
Proof.
  destruct x, y; ins; desf;
    (erewrite mk_slist_eq in Heq; 
      [|intros; rewrite in_app_iff at 1; rewrite in_app_iff; apply or_comm]); 
    desf.
Qed.

Lemma prot_plusA :
  forall x y xy (P1: prot_plus x y = Some xy) z xyz 
                (P2: prot_plus xy z = Some xyz),
    exists yz, prot_plus y z = Some yz /\ prot_plus x yz = Some xyz.
Proof.
  destruct x; ins; desf; eauto; try solve [eexists,; ins; desf].
   simpl in P2; desf; try solve[eexists,;ins;desf].
  cut (exists s5, mk_slist tau (SL_list s0 ++ SL_list s2) = Some s5 /\
                  mk_slist tau (SL_list s ++ SL_list s5) = Some s3).
    by clear; ins; desf; eexists,; ins; desf.
  unfold mk_slist in Heq, Heq0; desf; ins; desf.
  unfold mk_slist at 1; desf; ins.
  2: destruct n; split; try red; ins; [by apply app_eq_nil in H; destruct s2; desf|].
  2: apply a1; repeat first [rewrite in_app_iff in * | rewrite In_sort in * | rewrite in_undup_iff in *]; tauto. 
  eexists,; try done.
  unfold mk_slist; desf; ins; [f_equal; apply slist_eq; simpl|].
  { eapply sorted_perm_eq with (cmp := dec_prot_trans tau); eauto.
     eby clear; red; unfold dec_prot_trans; ins; desf; destruct n; eapply prot_trans_trans.
     eby clear; red; unfold dec_prot_trans; ins; desf; eapply prot_trans_ord.
    eapply NoDup_Permutation; eauto; clear; ins.
      repeat first [rewrite in_undup_iff | rewrite In_sort | rewrite in_app_iff ]; tauto.
    eapply Sorted_undup, sort_sorted.
  by clear - a5; desf; ins; specialize (a0 _ _ H H0); unfold dec_prot_trans; desf; eauto.
   eby clear; red; unfold dec_prot_trans; ins; desf; destruct n; eapply prot_trans_trans.
    eapply Sorted_undup, sort_sorted.
   by clear - a1; desf; ins; specialize (a1 _ _ H H0); unfold dec_prot_trans; desf; eauto.
   eby clear; red; unfold dec_prot_trans; ins; desf; destruct n; eapply prot_trans_trans.
  }
  destruct n; split; try red; ins; [by apply app_eq_nil in H; destruct s; desf|].
  apply a1; repeat first [rewrite in_app_iff in * | rewrite In_sort in * 
                          | rewrite in_undup_iff in *]; tauto. 
Qed.

(** ACIK properties of ghost resource composition *)

Lemma gres_ext f g : (forall n, gres_fun f n = gres_fun g n) -> f = g.
Proof.
  ins; apply functional_extensionality in H.
  destruct f, g; ins; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma gres_plusC x y : gres_plus x y = gres_plus y x.
Proof.
  intros; unfold gres_plus; repeat des_eqrefl; f_equal;
    try first [apply gres_ext; ins | exfalso];
    rewrite lift_plusC in EQ; try congruence; eauto using elem_gres_plusC.
Qed.

Lemma gres_plusI x : gres_plus gres_emp x = Some x.
Proof.
  unfold gres_plus, gres_emp; ins; des_eqrefl; f_equal;
  try first [apply gres_ext; ins | exfalso];
  rewrite lift_plus_emp_l in EQ; try congruence; apply elem_gres_plusI. 
Qed.

Lemma gres_plus_emp_r x : gres_plus x gres_emp = Some x.
Proof.
  rewrite gres_plusC; apply gres_plusI.
Qed.

Lemma gres_plusA : 
  forall x y xy (P1: gres_plus x y = Some xy) z xyz 
                (P2: gres_plus xy z = Some xyz),
    exists yz, gres_plus x yz = Some xyz /\ gres_plus y z = Some yz. 
Proof.
  unfold gres_plus; ins; repeat des_eqrefl; desf; ins.
    eexists,; ins; ins; des_eqrefl; f_equal;
    try first [apply gres_ext; ins | exfalso];
    symmetry in EQ0, EQ1, EQ, EQ2;
    eapply lift_plusA in EQ; eauto; ins; desf; eauto using elem_gres_plusA;
    by eapply elem_gres_plusA in P2; desf; eauto.
  exfalso; symmetry in EQ0, EQ1, EQ;
  eapply lift_plusA in EQ; eauto; ins; desf; eauto using elem_gres_plusA.
  by eapply elem_gres_plusA in P2; desf; eauto.
Qed.

Lemma gres_plusA2 : 
  forall x y xy (P1: gres_plus x y = Some xy) z (P2: gres_plus xy z = None),
    gres_plus y z = None
    \/ exists yz, gres_plus x yz = None /\ gres_plus y z = Some yz.
Proof.
  unfold gres_plus; ins; repeat des_eqrefl; desf; ins.
  symmetry in EQ, EQ0; eapply lift_plusA2 in EQ; eauto; desf;
   [left | right | ins]; try des_eqrefl; ins;
    try (exfalso; congruence).
  eexists,; ins; ins; des_eqrefl; ins; exfalso; congruence.
  eapply elem_gres_plusA2 in P2; eauto; desf; eauto.
Qed.

Lemma gres_eq_empD x y : gres_plus x y = Some gres_emp -> x = gres_emp.
Proof.
  unfold gres_plus, gres_emp; ins; des_eqrefl; apply gres_ext; ins; desf.
  inv_lplus n; eapply elem_gres_eq_empD in EQ; ins.
Qed.

Lemma gres_strip_emp : gres_strip gres_emp = gres_emp.
Proof.
  by apply gres_ext; ins; apply elem_gres_strip_emp.
Qed.

Lemma gres_strip_plus x y xy :
  gres_plus x y = Some xy ->
  gres_plus (gres_strip x) (gres_strip y) = Some (gres_strip xy).
Proof.
  unfold gres_plus; ins; repeat (des_eqrefl; ins); f_equal; 
    first [apply gres_ext | exfalso]; ins; desf; ins;
  inv_lplus2 n; desf; apply elem_gres_strip_plus in EQ; congruence.
Qed.

Lemma gres_strip_idemp x : gres_strip (gres_strip x) = gres_strip x.
Proof.
  by apply gres_ext; ins; apply elem_gres_strip_idemp.
Qed.

Lemma gres_plus_strip_idemp x : gres_plus x (gres_strip x) = Some x.
Proof.
  unfold gres_plus; des_eqrefl; f_equal; try apply gres_ext; ins; 
  inv_lplus2 n; rewrite elem_gres_plus_strip_idemp in *; desf.
Qed.


Lemma gres_strip_max x y xy z : 
  gres_plus x y = Some xy ->
  gres_plus xy z = Some x ->
  gres_strip y = y.
Proof.
  unfold gres_plus; ins; apply gres_ext; ins.
  ins; repeat (des_eqrefl; ins); desf.
  destruct x, y, z; ins; desf; inv_lplus n.
  eby eapply elem_gres_strip_max.
Qed.

(*Lemma gres_strip_max2 x y : gres_plus x y = Some x -> gres_strip y = y.
Proof.
  unfold gres_plus; ins; repeat (des_eqrefl; ins); f_equal; desf.
  destruct x; desf; try apply gres_ext; ins; inv_lplus n;
  eauto using elem_gres_strip_max.
Qed.*)

Lemma gres_pcancel :
  forall g g1 g2 gg,
    gres_plus g g1 = Some gg ->
    gres_plus g g2 = Some gg -> 
    gres_plus (gres_strip g) g1 = gres_plus (gres_strip g) g2.
Proof.
  unfold gres_plus; ins; repeat (des_eqrefl; ins); f_equal;
  first [ apply gres_ext; ins | exfalso ]; desf; inv_lplus2 n;
  erewrite elem_gres_pcancel with (g2 := g2 n) in EQ1; eauto; congruence.
Qed.

Arguments gres_plusA   [x y xy] P1 [z xyz] P2.
Arguments gres_plusA2  [x y xy] P1 [z] P2.
Arguments gres_eq_empD [x y] P. 
Arguments gres_pcancel [g g1 g2 gg] P1 P2. 

(** ACIZ properties of resource composition *)

Lemma res_plus_undef r : res_plus r Rundef = Rundef.
Proof. by destruct r. Qed.

Lemma res_plusC x y : res_plus x y = res_plus y x.
Proof. 
  unfold res_plus; desf; rewrite lift_plusC in Heq; auto using prot_plusC;
    rewrite ?(gres_plusC gres) in *; desf.
  f_equal; ins; extensionality a; apply orbC. 
Qed.

Lemma res_plus_emp_l x : res_plus res_emp x = x.
Proof. 
  repeat split; ins; desf; try rewrite lift_plus_emp_l, gres_plusI in *; desf.
  by rewrite lift_plus_noneD in *; desf.
Qed.

Lemma res_plus_emp_r x : res_plus x res_emp = x.
Proof. by rewrite res_plusC, res_plus_emp_l. Qed.

Lemma res_plusA x y z : res_plus (res_plus x y) z = res_plus x (res_plus y z).
Proof.
  unfold res_plus at 3 4; desf.
  by unfold res_plus; desf.
  by unfold res_plus; desf; exploit (@gres_plusA gres); eauto; ins; desf.
  by unfold res_plus; desf; exploit (lift_plusA _ prot_plusA pmap); 
     eauto; ins; desf.
  {
    rewrite gres_plusC, (lift_plusC _ prot_plusC) in * |-.
    exploit (lift_plusA _ prot_plusA pmap2); eauto.
    exploit (@gres_plusA gres2); eauto.
    intros; desf; rewrite gres_plusC, (lift_plusC _ prot_plusC) in * |-.
    unfold res_plus; desf; desf.
    by f_equal; ins; extensionality a; apply eq_sym, orbA. 
  }
  by unfold res_plus; desf; exploit (@gres_plusA gres); eauto; ins; desf.
  by unfold res_plus; desf; exploit (lift_plusA _ prot_plusA pmap); eauto; ins; desf.
Qed.

Lemma res_plusAC x y z : res_plus y (res_plus x z) = res_plus x (res_plus y z).
Proof.
  rewrite <- !res_plusA; f_equal; apply res_plusC.
Qed.

Ltac kundef :=
  match goal with 
    H: _ <> Rundef |- _ <> Rundef => 
      let X := fresh in intro X; apply H; 
      clear - X; rewrite X, ?res_plus_undef; simpl; done
  end.

(** The composition preorder on resources *)

Definition res_lt r r' := exists r_rem, res_plus r r_rem = r'.

Lemma res_lt_refl r : res_lt r r.
Proof. red; eauto using res_plus_emp_r. Qed.

Lemma res_lt_trans x y z : res_lt x y -> res_lt y z -> res_lt x z.
Proof. 
  unfold res_lt; ins; desf; rewrite res_plusA; vauto.
Qed.

(******************************************************************************)
(** ** 2.1.4. Resource stripping *) 
(******************************************************************************)

Definition prot_strip (pi : protocol) := 
  match pi with
    | p_at _ _ => pi
    | _ => p_bot
  end.

Definition res_strip (r : resource) :=
  match r with
    | Rundef => Rundef
    | Rdef p g e => Rdef (fun loc => prot_strip (p loc)) (gres_strip g) e
  end.

Lemma res_strip_idemp r : res_strip (res_strip r) = res_strip r.
Proof.
  destruct r; ins; f_equal; auto using gres_strip_idemp.
  by extensionality loc; destruct (pmap loc); ins.
Qed.

Lemma or_same P : P \/ P <-> P.
Proof. tauto. Qed.

Lemma mk_slist_idemp tau s : 
  mk_slist tau (SL_list s ++ SL_list s) = Some s. 
Proof.
  replace (Some s) with (mk_slist tau (SL_list s)).  
     apply mk_slist_eq; ins; rewrite in_app_iff; tauto.
  unfold mk_slist; destruct s; desf; ins; f_equal; try apply slist_eq; ins.
  2: destruct n; split; ins; eapply Sorted_total; try red; eauto using prot_trans_refl.
  desf; eapply sorted_perm_eq; eauto using prot_trans_trans; try apply prot_trans_ord.
    by eapply NoDup_Permutation; eauto; ins; rewrite in_undup_iff, In_sort.
  by apply Sorted_undup, Sorted_prot.
Qed.

Lemma res_plus_strip_idemp r : res_plus r (res_strip r) = r.
Proof.
  destruct r; ins; desf;
    try rewrite gres_plusC, gres_plusI in Heq0; desf; f_equal;
    try rewrite gres_plus_strip_idemp in *; try congruence.

  extensionality l; eapply lift_plus_some_inv with (l := l) in Heq.
  destruct (pmap l); ins; desf;  rewrite mk_slist_idemp in *; desf. 

  by extensionality sigma; apply orbb. 

  apply lift_plus_none_inv in Heq; desf.
  destruct (pmap l); ins; desf;  rewrite mk_slist_idemp in *; desf. 
Qed.
 
Lemma res_strip_plus r r' :
  res_plus r r' <> Rundef ->
  res_strip (res_plus r r') = res_plus (res_strip r) (res_strip r').
Proof.
  destruct r, r'; ins; desf; ins; try rewrite gres_plusI in *; desf; f_equal;
    try (by apply gres_strip_plus in Heq0; desf).
    extensionality l; inv_lplus l.
    by destruct (pmap l), (pmap0 l); ins; destruct (p l), (p0 l); ins; desf. 
  by exfalso; inv_lplus2 l; destruct (pmap l), (pmap0 l); ins; desf.
Qed.

Lemma res_escr_plusE r r' sigma :
  res_escr (res_plus r r') sigma -> res_escr r sigma \/ res_escr r' sigma.
Proof.
  unfold res_plus; desf; ins; destruct (escr sigma); vauto.
Qed.

Lemma res_escr_plusI r r' sigma :
  res_plus r r' <> Rundef ->
  res_escr r sigma \/ res_escr r' sigma -> res_escr (res_plus r r') sigma.
Proof.
  unfold res_plus; desf; ins; desf; apply orbT. 
Qed.

Lemma res_escr_plus r r' (D: res_plus r r' <> Rundef) sigma :
  res_escr (res_plus r r') sigma <-> res_escr r sigma \/ res_escr r' sigma.
Proof.
  split; eauto using res_escr_plusI, res_escr_plusE.
Qed.  


(******************************************************************************)
(** ** 2.1.5. Propositions *)
(******************************************************************************)

Section Propositions.

Record RProp :=
  { RP_car :> resource -> Prop ;
    RP_def : ~ RP_car Rundef ;
    RP_ext : forall r (IN: RP_car r) r' (DEF: res_plus r r' <> Rundef), 
      RP_car (res_plus r r') }.  

Local Obligation Tactic := 
  (try red; intros; desf; 
   repeat match goal with r : RProp |- _ => destruct r end; 
   ins; eauto).

Program Definition Rtrue := {| RP_car := fun r => r <> Rundef |}.
Program Definition Rfalse := {| RP_car := fun _ => False |}.

Program Definition Rpure P := {| RP_car := fun r => r <> Rundef /\ P |}.

Program Definition Rsingl (r: resource) :=
  {| RP_car r'' := r'' <> Rundef /\ exists r', r'' = res_plus r r' |}.
Next Obligation. eauto using res_plusA. Qed.

Program Definition Rstar (P1 P2: RProp) :=
  {| RP_car r := r <> Rundef /\ exists r1 r2, P1 r1 /\ P2 r2 /\ r = res_plus r1 r2 |}.
Next Obligation. 
  split; ins.
  eexists r1, (res_plus r2 r'); repeat split; eauto using res_plusA.
  apply RP_ext0; try done.
  by intro X; rewrite res_plusA, X, res_plusC in DEF.
Qed.

Program Definition Rconj (P Q : RProp) := 
  {| RP_car := fun r => P r /\ Q r |}.

Program Definition Rdisj (P Q : RProp) := 
  {| RP_car := fun r => P r \/ Q r |}.

Program Definition Rimpl (P Q : RProp) := 
  {| RP_car := fun r => r <> Rundef /\ 
                        (forall r', P (res_plus r r') -> Q (res_plus r r')) |}.
Next Obligation. split; ins; rewrite res_plusA in *; eauto. Qed. 

Program Definition Rforall A (PP : A -> RProp) := 
  {| RP_car := fun r => r <> Rundef /\ forall x, PP x r |}.
Next Obligation. 
  split; ins; specialize (IN0 x); destruct (PP x); ins; eauto.
Qed.

Program Definition Rexists A (PP : A -> RProp) := 
  {| RP_car := fun r => r <> Rundef /\ exists x, PP x r |}.
Next Obligation. split; ins; exists x; destruct PP; ins; eauto. Qed.

Program Definition Rbox (P: RProp) := 
  {| RP_car := fun r => P (res_strip r) |}.
Next Obligation.
  rewrite res_strip_plus; ins.
  eapply RP_ext0; eauto.
  rewrite <- res_strip_plus; 
  destruct (res_plus r r'); ins.
Qed.

Program Definition Runinit l := 
  {| RP_car := fun r => res_get r l = p_uninit |}.
Next Obligation.
  destruct r; ins; desf; inv_lplus l; rewrite IN in *; ins; desf.  
Qed.

Program Definition Rpt l v :=
  {| RP_car := fun r => res_get r l = p_na v |}.
Next Obligation.
  destruct r; ins; desf; inv_lplus l; rewrite IN in *; ins; desf.  
Qed.

Definition Rgres g :=
  Rsingl (Rdef (fun _ => p_bot) g (fun _ => false)).

Program Definition Rescr sigma :=
  {| RP_car := fun r => res_escr r sigma |}.
Next Obligation. destruct r, r'; ins; desf; vauto. Qed.

End Propositions.


(******************************************************************************)
(** ** 2.1.6. Escrow and protocol type interpretations *)
(******************************************************************************)

Parameter interp_prot : protTy -> state -> nat -> RProp.

Parameter interp_escr : escrowTy -> RProp * RProp.

Axiom interp_escrOK :
  forall sigma r, ~ Rstar (fst (interp_escr sigma)) (fst (interp_escr sigma)) r.


(******************************************************************************)
(******************************************************************************)
(** * 2.2. Local safety *)
(******************************************************************************)
(******************************************************************************)

(******************************************************************************)
(** ** 2.2.1.  Protocols *)
(******************************************************************************)

Definition env_move r loc val rX :=
  exists tau Srf, 
  << DEF: res_plus r rX <> Rundef >> /\
  << TOK: prot_trans_ok (res_get r loc) (res_get rX loc) >> /\
  << GRF: res_get rX loc = p_at tau Srf >> /\
  << SP : interp_prot tau (get_state Srf) val rX >>.

Definition extends_by_one s pi pi' :=
  (exists tau S S',
    pi = p_at tau S /\ pi' = p_at tau S' /\
    (forall s0, In s0 (SL_list S') <-> s0 = s \/ In s0 (SL_list S)) /\
    (forall s0 (SIN: In s0 (SL_list S)), prot_trans tau s0 s)) \/
  (exists tau, 
    pi = p_uninit /\ pi' = p_at tau (mk_slist_singl tau s)).

(******************************************************************************)
(** ** 2.2.2. Rely/guarantee *)
(******************************************************************************)

Require Import c11. 

Definition rely (r : resource) (alpha : act) (r': resource) :=
  match alpha with
    | Askip           => r' = r /\ r' <> Rundef
    | Aalloc l n      => r' = r /\ r' <> Rundef 
    | Aload typ l v =>
         typ = RATna /\ (exists pmap gres escr, 
                           r = Rdef pmap gres escr /\
                           r' = r /\ (forall v', pmap l = p_na v' -> v' = v))
       \/ typ = RATrlx /\ r' = r /\ r' <> Rundef 
       \/ is_acquire_rtyp typ /\ 
           exists rY, 
             (forall tau S, res_get r l = p_at tau S -> env_move r l v rY) /\ 
             r' = res_plus r (res_strip rY) /\ r' <> Rundef
    | Astore typ l _ =>
        r' = r /\ r' <> Rundef /\
        (forall (REL: is_release_wtyp typ) tau S, res_get r l = p_at tau S -> 
          exists v' rY, env_move r l v' rY)
    | Armw typ l v _ =>
      exists rY, 
        (forall (REL: is_release_utyp typ) (ACQ: is_acquire_utyp typ) tau S, 
          res_get r l = p_at tau S -> env_move r l v rY) /\ 
        r' = res_plus r rY /\ r' <> Rundef
  end.

Definition atomic_guar (r : resource) l v (r_sb r_rf: resource) :=
  exists tau s S',
    << PI : interp_prot tau s v r_rf >> /\
    << REQ: res_plus r_sb r_rf = res_upd r l (p_at tau S') >> /\
    << Gsb: res_get r_sb l = p_at tau S' >> /\
    << Grf: res_get r_rf l = p_at tau S' >> /\
    (<< GET: res_get r l = p_uninit >> /\ 
     << SEQ: SL_list S' = s :: nil >>
    \/ exists S, 
         << GET : res_get r l = p_at tau S >> /\ 
         << LAST: forall s0, In s0 (SL_list S) -> prot_trans tau s0 s >> /\
         << SEQ : forall s0, In s0 (SL_list S') <-> s0 = s \/ In s0 (SL_list S) >>).

Definition guar (r_pre r : resource) (alpha : act) (r_sb r_rf: resource) :=
  match alpha with
    | Askip           => r_sb = r /\ r_rf = res_emp /\ r <> Rundef
    | Aalloc l n      => r_sb = res_mupd r l n p_uninit
                         /\ r_rf = res_emp /\ r <> Rundef
    | Aload typ l _   => r_sb = r /\ r_rf = res_emp /\
      ((exists v, res_get r l = p_na v)
       \/ is_acquire_rtyp typ /\ (exists tau S, res_get r l = p_at tau S))
    | Astore typ l v =>
          typ = WATna /\ r_sb = res_upd r l (p_na v) /\ r_rf = res_emp /\ 
          (res_get r l = p_uninit \/ exists v, res_get r l = p_na v)
       \/ << ISREL: is_release_wtyp typ >> /\
          << AG: atomic_guar r l v r_sb r_rf >> /\
          << NB: res_get r_pre l <> p_bot >> /\
          << WG: forall v' rE (RELY: env_move r_pre l v' rE), 
                 prot_trans_ok (res_get rE l) (res_get r_rf l) >>
    | Armw typ l _ v      =>
        << ISACQ: is_acquire_utyp typ >> /\
        << ISREL: is_release_utyp typ >> /\
        << AG: atomic_guar r l v r_sb r_rf >> /\
        << NB: res_get r_pre l <> p_bot >> /\
        << WG: exists tau S, res_get r l = p_at tau S >>
  end.

(******************************************************************************)
(** ** 2.2.3. Ghost moves *)
(******************************************************************************)

Inductive ghost_move (r : resource) (Q : RProp) : Prop :=
  | GM_base (RINS: Q r) 
  | GM_trans P (GM: ghost_move r P) 
       (GM': forall r' (SAT: P r'), ghost_move r' Q)
  | GM_extend gone 
       (PEQ : forall l r'' pmap g g' escr, 
               gres_plus (gres_one l gone) g = Some g' ->
               res_plus r (Rdef pmap g' escr) = r'' ->
               r'' <> Rundef -> 
               Q r'')
  | GM_update p g e (REQ: r = Rdef p g e) g'
       (GDEF : forall gF, gres_plus g gF = None <-> gres_plus g' gF = None)
       (IN'  : Q (Rdef p g' e))
  | GM_eintro sigma r1 r2 P1 P2
       (ESCR: interp_escr sigma = (P1 , P2))
       (REQ : r = res_plus r1 r2) (RDEF: r <> Rundef)
       (RINS: P2 r2)
       (IN' : Q (res_plus r1 (Rdef (fun _ => p_bot) gres_emp 
                                   (fun s => s == sigma))))
  | GM_eelim sigma P1 P2 r0 rE
       (ESCR: interp_escr sigma = (P1 , P2))
       (REQ : r = res_plus rE r0) (RDEF: r <> Rundef)
       (VINS: P1 rE)
       (EIN : res_escr r sigma)
       (IN' : forall r'', Rstar (Rsingl r0) P2 r'' -> Q r'').

Lemma GM_weaken_post :
  forall r Q (GM: ghost_move r Q) (Q' : RProp)
         (IMP: forall r, Q r -> Q' r),
  ghost_move r Q'.
Proof.
  induction 1; ins; eauto using ghost_move. 
Qed.


(******************************************************************************)
(** ** 2.2.4. Local safety *)
(******************************************************************************)

Definition write_prot_equiv alpha r_pre r' :=
  match alpha with
    | Aalloc l n =>
          forall loc, l <= loc <= l + n -> res_get r' loc = p_bot
    | Astore typ l _ =>
        forall (ISREL: is_release_wtyp typ) tau S_pre S,
          res_get r_pre l = p_at tau S_pre ->
          res_get r' l = p_at tau S ->
          exists r_env v_env, 
            (exists Sxxx, res_get r_env l = p_at tau Sxxx
                  /\ slist_lt S Sxxx) /\
            env_move r_pre l v_env r_env
    | Armw typ l _ _ =>
        forall (ISACQ: is_acquire_utyp typ) (ISREL: is_release_utyp typ) tau S,
          res_get r_pre l = p_at tau S ->
          exists Su,
            res_get r' l = p_at tau Su /\ get_state S = get_state Su
    | _ => True
  end.

Fixpoint safe n (e : exp) PP r :=
  match n with 0 => True | S n =>
    << END: forall val, e = Evalue val -> ghost_move r (PP val) >> /\
    << FORK: forall K (CTX: eval_ctx K) e', e = K (Efork e') -> 
         exists r1 r2, r = res_plus r1 r2 /\ 
           safe n (K (Evalue 0)) PP r1 /\ 
           safe n e' (fun _ => Rtrue) r2 >> /\
    << STEP: forall alpha e' (STEP: exp_step e alpha e') rF r_pre
               (RELY : rely (res_plus r rF) alpha r_pre),
         exists P', 
           ghost_move r_pre P' /\
           forall r' (IN': P' r') (WPE : write_prot_equiv alpha r_pre r'), 
           exists r_post rx, 
             guar r_pre r' alpha (res_plus r_post rF) rx /\ 
             safe n e' PP r_post >>
  end.
